from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from vispy.geometry import MeshData
from vispy.scene import visuals
from vispy.visuals.filters import WireframeFilter

from whitecanvas.backend import _not_implemented


class Mesh3D(visuals.Mesh):
    mesh_data: MeshData

    def __init__(self, verts: NDArray[np.floating], faces: NDArray[np.intp]):
        self.wireframe_filter = WireframeFilter()
        super().__init__(vertices=verts, faces=faces)
        self.attach(self.wireframe_filter)
        self.unfreeze()

    def _plt_get_data(self):
        return self.mesh_data.get_vertices(), self.mesh_data.get_faces()

    def _plt_set_data(self, verts, faces):
        self.mesh_data.set_vertices(verts)
        self.mesh_data.set_faces(faces)
        self._update()

    def _plt_get_face_color(self) -> NDArray[np.float32]:
        return self.mesh_data.get_face_colors()

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        self.mesh_data.set_face_colors(color)
        self._update()

    _plt_get_face_hatch, _plt_set_face_hatch = _not_implemented.face_hatches()

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return np.array(self.wireframe_filter.color)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self.wireframe_filter.color = color
        self.update()

    def _plt_get_edge_width(self) -> float:
        return self.wireframe_filter.width

    def _plt_set_edge_width(self, width: float):
        self.wireframe_filter.width = width
        self.update()

    _plt_get_edge_style, _plt_set_edge_style = _not_implemented.edge_style()

    def _update(self):
        self._bounds = self.mesh_data.get_bounds()
        self.mesh_data_changed()

    def _plt_get_ndata(self) -> int:
        return self.mesh_data.get_faces().shape[0]



from __future__ import annotations

from typing import Callable, Iterator

import numpy as np
from bokeh import events as bk_events
from bokeh import layouts as bk_layouts
from bokeh import models as bk_models
from bokeh import plotting as bk_plotting
from bokeh.io import output_notebook

from whitecanvas import protocols
from whitecanvas.backend.bokeh._base import BokehLayer
from whitecanvas.backend.bokeh._labels import (
    Title,
    XAxis,
    XLabel,
    XTicks,
    YAxis,
    YLabel,
    YTicks,
)
from whitecanvas.backend.bokeh._legend import make_sample_item
from whitecanvas.layers._legend import LegendItem, LegendItemCollection
from whitecanvas.types import (
    Location,
    Modifier,
    MouseButton,
    MouseEvent,
    MouseEventType,
)
from whitecanvas.utils.normalize import arr_color, hex_color


def _prep_plot(width=400, height=300) -> bk_plotting.figure:
    plot = bk_plotting.figure(
        width=width,
        height=height,
        tooltips="@hovertexts",
    )
    plot.title.align = "center"
    return plot


SECOND_X = "second-x"
SECOND_Y = "second-y"


@protocols.check_protocol(protocols.CanvasProtocol)
class Canvas:
    def __init__(
        self,
        plot: bk_models.Plot,
        second_x: bool = False,
        second_y: bool = False,
    ):
        assert isinstance(plot, bk_models.Plot)
        self._plot = plot
        self._xaxis = XAxis(self)
        self._yaxis = YAxis(self)
        self._title = Title(self)
        self._xlabel = XLabel(self)
        self._ylabel = YLabel(self)
        self._xticks = XTicks(self)
        self._yticks = YTicks(self)
        self._mouse_button: MouseButton = MouseButton.NONE
        self._second_x = second_x
        self._second_y = second_y

    def _set_mouse_down(self, event):
        self._mouse_button = event

    def _get_xaxis(self):
        if not self._second_x:
            return self._plot.xaxis[0]
        return self._plot.xaxis[1]

    def _get_yaxis(self):
        if not self._second_y:
            return self._plot.yaxis[0]
        return self._plot.yaxis[1]

    def _get_xrange(self):
        if not self._second_x:
            return self._plot.x_range
        return self._plot.extra_x_ranges[SECOND_X]

    def _get_yrange(self):
        if not self._second_y:
            return self._plot.y_range
        return self._plot.extra_y_ranges[SECOND_Y]

    def _plt_get_native(self):
        return self._plot

    def _plt_get_title(self):
        return self._title

    def _plt_get_xaxis(self):
        return self._xaxis

    def _plt_get_yaxis(self):
        return self._yaxis

    def _plt_get_xlabel(self):
        return self._xlabel

    def _plt_get_ylabel(self):
        return self._ylabel

    def _plt_get_xticks(self):
        return self._xticks

    def _plt_get_yticks(self):
        return self._yticks

    def _bokeh_renderers(self) -> list[bk_models.GlyphRenderer]:
        return self._plot.renderers

    def _plt_reorder_layers(self, layers: list[BokehLayer]):
        # NOTE: if plot has second_x or second_y, renderers will have more than layers
        model_to_idx_map = {id(layer._model): i for i, layer in enumerate(layers)}
        existing_glyphs = {id(layer._model) for layer in layers}
        renderers = self._bokeh_renderers()
        idx_converter = []
        for i, renderer in enumerate(renderers):
            if id(renderer.glyph) in existing_glyphs:
                idx_converter.append(i)
        renderers_sorted = []
        for r in renderers:
            idx = model_to_idx_map.get(id(r.glyph))
            if idx is not None:
                renderers_sorted.append(renderers[idx_converter[idx]])
            else:
                renderers_sorted.append(r)
        self._plot.renderers = renderers_sorted

    def _plt_get_aspect_ratio(self) -> float | None:
        return self._plot.aspect_ratio

    def _plt_set_aspect_ratio(self, ratio: float | None):
        self._plot.aspect_ratio = ratio

    def _plt_add_layer(self, layer: BokehLayer):
        if self._second_y:
            self._plot.add_glyph(layer._data, layer._model, y_range_name=SECOND_Y)
        elif self._second_x:
            self._plot.add_glyph(layer._data, layer._model, x_range_name=SECOND_X)
        else:
            self._plot.add_glyph(layer._data, layer._model)

    def _plt_remove_layer(self, layer: BokehLayer):
        """Remove layer from the canvas"""
        idx = -1
        for i, renderer in enumerate(self._bokeh_renderers()):
            if renderer.glyph == layer._model:
                idx = i
                del self._plot.renderers[idx]
                break
        else:
            raise ValueError(f"Layer {layer} not found")

    def _plt_get_visible(self) -> bool:
        """Get visibility of canvas"""
        return self._plot.visible

    def _plt_set_visible(self, visible: bool):
        """Set visibility of canvas"""
        self._plot.visible = visible

    def _plt_connect_mouse_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(event: bk_events.Press):
            self._set_mouse_down(MouseButton.LEFT)
            ev = MouseEvent(
                button=MouseButton.LEFT,
                modifiers=_translate_modifiers(event.modifiers),
                pos=(event.x, event.y),
                type=MouseEventType.PRESS,
            )
            callback(ev)

        self._plot.on_event(bk_events.Tap, _cb)
        self._plot.on_event(bk_events.Press, _cb)

    def _plt_connect_mouse_drag(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(event: bk_events.MouseMove):
            ev = MouseEvent(
                button=self._mouse_button,
                modifiers=_translate_modifiers(event.modifiers),
                pos=(event.x, event.y),
                type=MouseEventType.MOVE,
            )
            callback(ev)

        self._plot.on_event(bk_events.MouseMove, _cb)

    def _plt_connect_mouse_double_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(event: bk_events.DoubleTap):
            ev = MouseEvent(
                button=MouseButton.LEFT,
                modifiers=_translate_modifiers(event.modifiers),
                pos=(event.x, event.y),
                type=MouseEventType.DOUBLE_CLICK,
            )
            callback(ev)

        self._plot.on_event(bk_events.DoubleTap, _cb)

    def _plt_connect_mouse_release(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(event: bk_events.PressUp):
            ev = MouseEvent(
                button=MouseButton.LEFT,
                modifiers=_translate_modifiers(event.modifiers),
                pos=(event.x, event.y),
                type=MouseEventType.RELEASE,
            )
            callback(ev)
            self._set_mouse_down(MouseButton.NONE)

        self._plot.on_event(bk_events.PressUp, _cb)
        self._plot.on_event(bk_events.Tap, _cb)
        self._plot.on_event(bk_events.PanEnd, _cb)

    def _plt_connect_xlim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        rng = self._plot.x_range
        rng.on_change(
            "start",
            lambda attr, old, new: callback((rng.start, rng.end)),  # noqa: ARG005
        )

    def _plt_connect_ylim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        rng = self._plot.y_range
        rng.on_change(
            "start",
            lambda attr, old, new: callback((rng.start, rng.end)),  # noqa: ARG005
        )

    def _plt_draw(self):
        pass

    def _plt_get_mouse_enabled(self):
        return self._plot.toolbar.active_drag is None

    def _plt_set_mouse_enabled(self, enabled: bool):
        self._plot.toolbar.active_drag = "auto" if enabled else None

    def _plt_twinx(self):
        self._plot.extra_y_ranges[SECOND_Y] = bk_models.DataRange1d()
        self._plot.add_layout(
            bk_models.LinearAxis(y_range_name=SECOND_Y),
            "right",
        )
        return Canvas(self._plot, second_y=True)

    def _plt_twiny(self):
        self._plot.extra_x_ranges[SECOND_X] = bk_models.DataRange1d()
        self._plot.add_layout(
            bk_models.LinearAxis(x_range_name=SECOND_X),
            "above",
        )
        return Canvas(self._plot, second_x=True)

    def _plt_make_legend(
        self,
        items: list[tuple[str, LegendItem]],
        anchor: Location = Location.TOP_RIGHT,
    ):
        bk_items = []
        bk_samples = []
        for label, item in items:
            if item is None:
                continue
            if isinstance(item, LegendItemCollection):
                for sub_label, sub_item in item.items:
                    sample = make_sample_item(sub_item)
                    if sample is not None:
                        bk_items.append((sub_label, sample))
                        bk_samples.extend(sample)
            else:
                sample = make_sample_item(item)
                if sample is not None:
                    bk_items.append((label, sample))
                    bk_samples.extend(sample)
        location, side = _LEGEND_LOCATIONS[anchor]
        legend = bk_models.Legend(items=bk_items, location=location)
        self._plot.add_layout(legend, side)
        self._plot.renderers.extend(bk_samples)


def _translate_modifiers(mod: bk_events.KeyModifiers | None) -> tuple[Modifier, ...]:
    if mod is None:
        return ()
    modifiers = []
    if mod["alt"]:
        modifiers.append(Modifier.ALT)
    if mod["shift"]:
        modifiers.append(Modifier.SHIFT)
    if mod["ctrl"]:
        modifiers.append(Modifier.CTRL)
    return tuple(modifiers)


# location and side
_LEGEND_LOCATIONS = {
    Location.TOP_LEFT: ("top_left", "center"),
    Location.TOP_CENTER: ("center_top", "center"),
    Location.TOP_RIGHT: ("top_right", "center"),
    Location.CENTER_LEFT: ("center_left", "center"),
    Location.CENTER: ("center_center", "center"),
    Location.CENTER_RIGHT: ("center_right", "center"),
    Location.BOTTOM_LEFT: ("bottom_left", "center"),
    Location.BOTTOM_CENTER: ("bottom_center", "center"),
    Location.BOTTOM_RIGHT: ("bottom_right", "center"),
    Location.LEFT_SIDE_TOP: ("top", "left"),
    Location.LEFT_SIDE_CENTER: ("center", "left"),
    Location.LEFT_SIDE_BOTTOM: ("bottom", "left"),
    Location.RIGHT_SIDE_TOP: ("top", "right"),
    Location.RIGHT_SIDE_CENTER: ("center", "right"),
    Location.RIGHT_SIDE_BOTTOM: ("bottom", "right"),
    Location.TOP_SIDE_LEFT: ("left", "above"),
    Location.TOP_SIDE_CENTER: ("center", "above"),
    Location.TOP_SIDE_RIGHT: ("right", "above"),
    Location.BOTTOM_SIDE_LEFT: ("left", "below"),
    Location.BOTTOM_SIDE_CENTER: ("center", "below"),
    Location.BOTTOM_SIDE_RIGHT: ("right", "below"),
}


@protocols.check_protocol(protocols.CanvasGridProtocol)
class CanvasGrid:
    def __init__(self, heights: list[float], widths: list[float], app: str = "default"):
        hsum = sum(heights)
        wsum = sum(widths)
        children = []
        for h in heights:
            row = []
            for w in widths:
                p = _prep_plot(width=int(w / wsum * 600), height=int(h / hsum * 600))
                p.visible = False
                wheel_zoom = bk_models.WheelZoomTool()
                p.add_tools(wheel_zoom)
                row.append(p)
            children.append(row)
        self._grid_plot: bk_layouts.GridPlot = bk_layouts.gridplot(
            children, sizing_mode="fixed"
        )
        self._widths = widths
        self._heights = heights
        self._width_total = wsum
        self._height_total = hsum
        self._app = app

    def _plt_add_canvas(self, row: int, col: int, rowspan: int, colspan: int) -> Canvas:
        for r0, c0, plot in self._iter_bokeh_subplots():
            if r0 == row and c0 == col:
                plot.visible = True
                return Canvas(plot)
        raise ValueError(f"Canvas at ({row}, {col}) not found")

    def _plt_show(self):
        if self._app == "notebook":
            output_notebook()
            bk_plotting.show(lambda doc: doc.add_root(self._grid_plot))
        else:
            bk_plotting.show(self._grid_plot)
            if self._grid_plot.document is None:
                bk_plotting.curdoc().add_root(self._grid_plot)

    def _plt_get_background_color(self):
        for _, _, child in self._iter_bokeh_subplots():
            return child.background_fill_color
        return np.ones(4, dtype=np.float32)

    def _plt_set_background_color(self, color):
        color = hex_color(color)
        for _, _, child in self._iter_bokeh_subplots():
            child.background_fill_color = color

    def _plt_screenshot(self):
        import io

        from bokeh.io import export_png

        with io.BytesIO() as buff:
            export_png(self._grid_plot, filename=buff)
            buff.seek(0)
            data = np.frombuffer(buff.getvalue(), dtype=np.uint8)
        w, h = self._grid_plot.width, self._grid_plot.height
        img = data.reshape((int(h), int(w), -1))
        return img

    def _plt_set_figsize(self, width: int, height: int):
        for r, c, child in self._iter_bokeh_subplots():
            child.height = int(self._heights[r] / self._height_total * height)
            child.width = int(self._widths[c] / self._width_total * width)
        self._grid_plot.width = width
        self._grid_plot.height = height

    def _plt_set_spacings(self, wspace: float, hspace: float):
        self._grid_plot.spacing = (int(wspace), int(hspace))

    def _iter_bokeh_subplots(self) -> Iterator[tuple[int, int, bk_plotting.figure]]:
        for child, r, c in self._grid_plot.children:
            yield r, c, child



import numpy as np

from whitecanvas import new_canvas


def main():
    canvas = new_canvas(backend="matplotlib:qt")
    x = np.arange(30)
    y = np.sin(x / 5)
    yerr = np.random.default_rng(12345).normal(size=30, scale=0.2)

    layer = (
        canvas
        .add_line(x, y, width=2, color="black")
        .with_markers(symbol="o", size=12)
        .with_yerr(yerr, capsize=0.5, color="black")
    )

    print("Layer of type", type(layer), "added.")  # noqa: T201

    canvas.show(block=True)

if __name__ == "__main__":
    main()



from __future__ import annotations

import re
import weakref
from typing import TYPE_CHECKING, Generic, TypeVar

import numpy as np

from whitecanvas._exceptions import ReferenceDeletedError
from whitecanvas.layers import Band, Bars, Layer
from whitecanvas.layers.group import LabeledBars, StemPlot
from whitecanvas.types import ArrayLike1D, ColorType, Hatch
from whitecanvas.utils.normalize import as_array_1d

if TYPE_CHECKING:
    from whitecanvas.canvas._base import CanvasBase

_C = TypeVar("_C", bound="CanvasBase")
_L = TypeVar("_L", bound=Layer)


class StackOverPlotter(Generic[_C, _L]):
    def __init__(self, canvas: _C, over: _L):
        self._canvas_ref = weakref.ref(canvas)
        self._relative_to_ref = weakref.ref(over)

    def _canvas(self) -> _C:
        canvas = self._canvas_ref()
        if canvas is None:
            raise ReferenceDeletedError("Canvas has been deleted.")
        return canvas

    def _layer(self) -> _L:
        layer = self._relative_to_ref()
        if layer is None:
            raise ReferenceDeletedError("Layer has been deleted.")
        return layer

    def add(
        self,
        ydata: ArrayLike1D,
        *,
        color: ColorType | None = None,
        hatch: str | Hatch = Hatch.SOLID,
        alpha: float = 1.0,
        name: str | None = None,
    ) -> _L:
        canvas = self._canvas()
        layer = self._layer()
        color = canvas._generate_colors(color)

        if name is None:
            if re.match(r".+\+\d", layer.name):
                stem, num = layer.name.rsplit("+", 1)
                name = f"{stem}+{int(num) + 1}"
            else:
                name = f"{layer.name}+1"

        # unwrap nested layers in a group
        if isinstance(layer, LabeledBars):
            layer = layer.bars

        if isinstance(layer, Bars):
            new_layer = Bars(
                layer.data.x,
                ydata,
                bottom=layer.top,
                orient=layer.orient,
                color=color,
                alpha=alpha,
                name=name,
                hatch=hatch,
                extent=layer.bar_width,
                backend=layer._backend_name,
            )
        elif isinstance(layer, Band):
            y0 = np.maximum(layer.data.y0, layer.data.y1)
            new_layer = Band(
                layer.data.x,
                y0,
                ydata + y0,
                orient=layer.orient,
                color=color,
                alpha=alpha,
                name=name,
                hatch=hatch,
                backend=layer._backend_name,
            )
        elif isinstance(layer, StemPlot):
            new_layer = StemPlot.from_arrays(
                layer.data.x,
                ydata,
                bottom=layer.top,
                name=name,
                orient=layer.orient,
                backend=layer._backend_name,
            )
        else:
            raise TypeError("Only Bars and Band are supported.")
        canvas.add_layer(new_layer)
        return new_layer

    def add_hist(
        self,
        data: ArrayLike1D,
        *,
        name: str | None = None,
        color: ColorType | None = None,
        hatch: str | Hatch = Hatch.SOLID,
        alpha: float = 1.0,
    ) -> _L:
        """
        Add data as histogram on top of the existing histogram.

        Parameters
        ----------
        data : array-like
            Data to be added as histogram.
        name : str, optional
            Name of the layer.
        color : color-like, optional
            Color of the layer face.
        alpha : float, default 1.0
            Alpha channel of the bars.
        hatch : str or FacePattern, default FacePattern.SOLID
            Hatch Pattern of the bar faces.

        Returns
        -------
        Bars
            The newly added histogram layer.
        """
        data = as_array_1d(data)
        canvas = self._canvas()
        layer = self._layer()
        color = canvas._generate_colors(color)
        if isinstance(layer, LabeledBars):
            layer = layer.bars

        if not isinstance(layer, Bars):
            raise TypeError(f"Can not stack histogram on {layer!r}.")
        if not layer._bar_type.startswith("histogram"):
            raise TypeError(f"{layer!r} is not histogram.")
        centers = layer.data.x
        density = layer._bar_type == "histogram-density"
        dx = layer.bar_width / 2
        bins = np.concatenate([[centers[0] - dx], centers + dx])
        counts, edges = np.histogram(
            data, bins, density=density, range=(bins.min(), bins.max())
        )
        new_layer = Bars(
            centers, counts, bottom=layer.top, extent=dx * 2, name=name,
            color=color, alpha=alpha, orient=layer.orient, hatch=hatch,
            backend=layer._backend_name,
        )  # fmt: skip
        new_layer._bar_type = layer._bar_type
        return canvas.add_layer(new_layer)



from __future__ import annotations

import inspect
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generator,
    Generic,
    TypeVar,
    overload,
)

from psygnal import throttled

from whitecanvas.types import MouseEvent

if TYPE_CHECKING:
    from typing_extensions import Self

_R = TypeVar("_R")
_S = TypeVar("_S")
_Y = TypeVar("_Y")

GeneratorFunction = Callable[..., Generator[_R, _S, _Y]]
_G = TypeVar("_G", bound=GeneratorFunction)
_F = TypeVar("_F", bound=Callable)


class _Slot(Generic[_G]):
    def __init__(self, slot: _G, msec: int = 0, leading: bool = True):
        self._slot_orig = slot
        if not inspect.isgeneratorfunction(slot):
            fn = _to_generator_function(slot)
        else:
            fn = slot
        if msec > 0:
            fn = throttled(fn, timeout=msec, leading=leading)
        self._slot = fn
        self._generator: Generator | None = None
        self._current_event: MouseEvent | None = None

    def next(self, ev: MouseEvent):
        """Advance the slot."""
        if self._generator is None:
            self._generator = self._slot(ev)
            self._current_event = ev
        try:
            self._current_event.update(ev)
            next(self._generator)
        except StopIteration:
            self._generator = None
            return


def _to_generator_function(f: Callable[[MouseEvent], Any | None]):
    def _f(ev: MouseEvent):
        yield f(ev)

    return _f


class _MouseSignalMixin:
    def __init__(self):
        self._slots: list[_Slot] = []
        self._name: str | None = None

    def __set_name__(self, owner: type[Any], name: str) -> None:
        if self._name is None:
            self._name = name

    def __get__(self, instance: Any, owner: type[Any] | None = None) -> Self:
        if instance is None:
            return self
        out = self._copy()
        setattr(instance, self._name, out)
        return out

    def _copy(self):
        new = self.__class__()
        new._name = self._name
        return new


class MouseMoveSignal(_MouseSignalMixin):
    @overload
    def connect(self, slot: _G, *, msec: int = 0, leading: bool = True) -> _G: ...

    @overload
    def connect(
        self, slot: None = None, *, msec: int = 0, leading: bool = True
    ) -> Callable[[_G], _G]: ...

    def connect(self, slot=None, *, msec: int = 0, leading: bool = True):
        """Connect a mouse move callback."""

        def _inner(slot: _F) -> _F:
            if not callable(slot):
                raise TypeError(f"Can only connect callable object, got {slot!r}")
            self._slots.append(_Slot(slot, msec=msec, leading=leading))
            return slot

        return _inner(slot) if slot is not None else _inner

    def disconnect(
        self, slot: GeneratorFunction | None = None, missing_ok: bool = True
    ) -> None:
        if slot is None:
            return self._slots.clear()
        i = -1
        for _i, _slot in enumerate(self._slots):
            if _slot._slot_orig is slot:
                i = _i
                break
        if i > 0:
            self._slots.pop(i)
        elif not missing_ok:
            raise ValueError(f"Slot {slot!r} not found")
        return

    def emit(self, ev: MouseEvent) -> None:
        """Emit the mouse event"""
        for slot in self._slots:
            slot.next(ev)


class MouseSignal(_MouseSignalMixin, Generic[_R]):
    def __init__(self, typ: type[_R]):
        super().__init__()
        self._typ = typ

    def connect(self, slot: _F | None = None) -> _F:
        if not callable(slot):
            raise TypeError(f"Can only connect callable object, got {slot!r}")
        self._slots.append(slot)
        return slot

    def disconnect(self, slot: Callable | None = None, missing_ok: bool = True) -> None:
        if slot is None:
            return self._slots.clear()
        i = -1
        for _i, _slot in enumerate(self._slots):
            if _slot is slot:
                i = _i
                break
        if i > 0:
            self._slots.pop(i)
        elif not missing_ok:
            raise ValueError(f"Slot {slot!r} not found")
        return

    def emit(self, *args) -> None:
        for slot in self._slots:
            slot(*args)

    def _copy(self):
        new = self.__class__(self._typ)
        new._name = self._name
        return new



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import matplotlib.transforms as mtransforms
import numpy as np
from matplotlib.collections import PathCollection
from numpy.typing import NDArray

from whitecanvas.backend.matplotlib._base import MplMouseEventsMixin, symbol_to_path
from whitecanvas.protocols import MarkersProtocol, check_protocol
from whitecanvas.types import Hatch, LineStyle, Symbol
from whitecanvas.utils.normalize import as_color_array
from whitecanvas.utils.type_check import is_real_number

if TYPE_CHECKING:
    from whitecanvas.backend.matplotlib.canvas import Canvas


@check_protocol(MarkersProtocol)
class Markers(PathCollection, MplMouseEventsMixin):
    def __init__(self, xdata, ydata):
        offsets = np.stack([xdata, ydata], axis=1)
        self._symbol = Symbol.CIRCLE
        super().__init__(
            (symbol_to_path(self._symbol),),
            sizes=[6] * len(offsets),
            offsets=offsets,
            picker=True,
        )
        self.set_transform(mtransforms.IdentityTransform())
        self._edge_styles = [LineStyle.SOLID] * len(offsets)
        MplMouseEventsMixin.__init__(self)

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        offsets = self.get_offsets()
        return offsets[:, 0], offsets[:, 1]

    def _plt_set_data(self, xdata, ydata):
        data = np.stack([xdata, ydata], axis=1)
        self.set_offsets(data)

    ##### HasSymbol protocol #####

    def _plt_get_symbol(self) -> Symbol:
        return self._symbol

    def _plt_set_symbol(self, symbol: Symbol):
        path = symbol_to_path(symbol)
        self.set_paths([path] * len(self.get_offsets()))
        self._symbol = symbol

    def _plt_get_symbol_size(self) -> NDArray[np.floating]:
        return np.sqrt(self.get_sizes())

    def _plt_set_symbol_size(self, size: float | NDArray[np.floating]):
        if is_real_number(size):
            size = np.full(len(self.get_offsets()), size)
        self.set_sizes(size**2)

    ##### HasFaces protocol #####
    def _plt_get_face_color(self) -> NDArray[np.float32]:
        return self.get_facecolor()

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, len(self.get_offsets()))
        self.set_facecolor(color)

    def _plt_get_face_hatch(self) -> Hatch:
        return [Hatch(self.get_hatch() or "")] * len(self.get_offsets())

    def _plt_set_face_hatch(self, pattern: Hatch | list[Hatch]):
        if not isinstance(pattern, Hatch):
            if len(set(pattern)) > 1:
                warnings.warn(
                    "matplotlib markers do not support multiple hatch patterns.",
                    UserWarning,
                    stacklevel=2,
                )
            pattern = pattern[0]
        if pattern is Hatch.SOLID:
            ptn = None
        else:
            ptn = pattern.value
        self.set_hatch(ptn)

    ##### HasEdges protocol #####

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return self.get_edgecolor()

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, len(self.get_offsets()))
        self.set_edgecolor(color)

    def _plt_get_edge_style(self) -> list[LineStyle]:
        return self._edge_styles

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        if isinstance(style, LineStyle):
            styles = [style.value] * len(self.get_offsets())
            self._edge_styles = [style] * len(self.get_offsets())
        else:
            styles = [s.value for s in style]
            self._edge_styles = style
        self.set_linestyle(styles)

    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        return self.get_linewidth()

    def _plt_set_edge_width(self, width: float | NDArray[np.floating]):
        if is_real_number(width):
            width = np.full(len(self.get_offsets()), width)
        self.set_linewidth(width)

    def post_add(self, canvas: Canvas):
        MplMouseEventsMixin.post_add(self, canvas)
        self.set_offset_transform(canvas._axes.transData)



from __future__ import annotations

import warnings

import numpy as np
from numpy.typing import NDArray
from vispy.scene import visuals

from whitecanvas.backend import _not_implemented
from whitecanvas.protocols import TextProtocol, check_protocol
from whitecanvas.types import Alignment, LineStyle
from whitecanvas.utils.normalize import as_color_array
from whitecanvas.utils.type_check import is_real_number


@check_protocol(TextProtocol)
class Texts(visuals.Text):
    def __init__(
        self, x: NDArray[np.floating], y: NDArray[np.floating], text: list[str]
    ):
        if x.size > 0:
            pos = np.stack([x, y, np.zeros(x.size)], axis=1)
        else:
            pos = np.array([[0, 0, 0]], dtype=np.float32)
            text = [""]
        super().__init__(text, pos=pos)
        self.unfreeze()
        self._alignment = Alignment.BOTTOM_LEFT
        # NOTE: vispy does not support empty text layer. Here we use "_is_empty" to
        # specify whether the layer is empty, and pretend to be empty.
        self._is_empty = x.size == 0

    def _plt_get_visible(self) -> bool:
        return self.visible

    def _plt_set_visible(self, visible: bool):
        self.visible = visible

    ##### TextProtocol #####

    def _plt_get_text(self) -> list[str]:
        if self._is_empty:
            return []
        return self.text

    def _plt_set_text(self, text: list[str]):
        if self._is_empty:
            self.text = [""]
        else:
            self.text = text

    def _plt_get_ndata(self) -> int:
        if self._is_empty:
            return 0
        return len(self.text)

    def _plt_get_text_color(self):
        return self.color.rgba

    def _plt_set_text_color(self, color):
        col = as_color_array(color, self._plt_get_ndata())
        if not self._is_empty:
            self.color = col

    def _plt_get_text_size(self) -> NDArray[np.floating]:
        return np.full(self._plt_get_ndata(), self.font_size)

    def _plt_set_text_size(self, size: float | NDArray[np.floating]):
        if is_real_number(size):
            self.font_size = size
        else:
            candidates = np.unique(size)
            if candidates.size == 1:
                self.font_size = candidates[0]
            elif candidates.size == 0:
                pass
            else:
                warnings.warn(
                    "vispy Text layer does not support different font sizes. Set to "
                    "the average size.",
                    UserWarning,
                    stacklevel=4,
                )
                self.font_size = np.mean(size)

    def _plt_get_text_position(
        self,
    ) -> tuple[NDArray[np.floating], NDArray[np.floating]]:
        if self._is_empty:
            return np.array([], dtype=np.float32), np.array([], dtype=np.float32)
        return self.pos[:, 0], self.pos[:, 1]

    def _plt_set_text_position(
        self, position: tuple[NDArray[np.floating], NDArray[np.floating]]
    ):
        x, y = position
        if x.size == 0:
            self.pos = np.array([[0, 0, 0]], dtype=np.float32)
        else:
            self.pos = np.stack([x, y, np.zeros(x.size)], axis=1)
        self._is_empty = x.size == 0

    def _plt_get_text_anchor(self) -> Alignment:
        return self._alignment

    def _plt_set_text_anchor(self, anc: Alignment):
        va, ha = anc.split()
        self.anchors = va.value, ha.value
        self._alignment = anc

    def _plt_get_text_rotation(self) -> NDArray[np.floating]:
        return -self.rotation  # the +/- is reversed compared to other backends

    def _plt_set_text_rotation(self, rotation: float | NDArray[np.floating]):
        if self._is_empty:
            if is_real_number(rotation):
                self.rotation = np.array([-rotation])
            else:
                if rotation.size != 0:
                    raise ValueError(f"zero text but got {rotation.size} inputs.")
        else:
            if is_real_number(rotation):
                rotation = np.full(self._plt_get_ndata(), rotation)
            self.rotation = -rotation

    def _plt_get_text_fontfamily(self) -> str:
        return self.face

    def _plt_set_text_fontfamily(self, fontfamily: str):
        self.face = fontfamily

    def _plt_get_face_color(self):
        return np.zeros((self._plt_get_ndata(), 4))

    def _plt_set_face_color(self, color):
        pass

    _plt_get_face_hatch, _plt_set_face_hatch = _not_implemented.face_hatches()

    def _plt_get_edge_color(self):
        return np.zeros((self._plt_get_ndata(), 4))

    def _plt_set_edge_color(self, color):
        pass

    def _plt_get_edge_width(self) -> float:
        return np.zeros(self._plt_get_ndata())

    def _plt_set_edge_width(self, width: float | NDArray[np.floating]):
        pass

    def _plt_get_edge_style(self) -> LineStyle:
        return [LineStyle.SOLID] * self._plt_get_ndata()

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        pass



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar, overload

import numpy as np
from numpy.typing import NDArray
from typing_extensions import Self

from whitecanvas import protocols, theme
from whitecanvas.backend import Backend, patch_dummy_backend
from whitecanvas.canvas import _namespaces as _ns
from whitecanvas.canvas._base import CanvasNDBase
from whitecanvas.layers import layer3d
from whitecanvas.layers._base import Layer
from whitecanvas.types import (
    ArrayLike1D,
    ColormapType,
    ColorType,
    Hatch,
    LineStyle,
    Orientation,
)
from whitecanvas.utils.normalize import as_array_1d, normalize_xyz

if TYPE_CHECKING:
    from whitecanvas.canvas import CanvasGrid

_L3D = TypeVar("_L3D", bound=layer3d.Layer3D)


class Canvas3DBase(CanvasNDBase):
    _CURRENT_INSTANCE: Canvas3DBase | None = None

    z = _ns.ZAxisNamespace()

    def __init__(self, *, palette: ColormapType | None = None):
        super().__init__(palette=palette)
        self._autoscale_enabled = True
        if not self._get_backend().is_dummy():
            self._init_canvas()

    def _init_canvas(self):
        # default colors and font
        if self._get_backend().name in ("matplotlib", "plotly"):
            _t = theme.get_theme()
            _ft = _t.font
            self.x.color = _t.foreground_color
            self.y.color = _t.foreground_color
            self.x.label.update(family=_ft.family, color=_ft.color, size=_ft.size)
            self.y.label.update(family=_ft.family, color=_ft.color, size=_ft.size)
            self.title.update(family=_ft.family, color=_ft.color, size=_ft.size)
            self.x.ticks.update(family=_ft.family, color=_ft.color, size=_ft.size)
            self.y.ticks.update(family=_ft.family, color=_ft.color, size=_ft.size)

            # lim changed events
            canvas = self._canvas()
            canvas._plt_connect_xlim_changed(self._emit_xlim_changed)
            canvas._plt_connect_ylim_changed(self._emit_ylim_changed)
            canvas._plt_connect_zlim_changed(self._emit_zlim_changed)

            if hasattr(canvas, "_plt_canvas_hook"):
                canvas._plt_canvas_hook(self)

        # connect layer events
        self.layers.events.inserted.connect(
            self._cb_inserted, unique=True, max_args=None
        )
        self.layers.events.removed.connect(self._cb_removed, unique=True, max_args=None)
        self.layers.events.reordered.connect(
            self._cb_reordered, unique=True, max_args=None
        )
        self.layers.events.connect(self._draw_canvas, unique=True, max_args=None)

    def _autoscale_for_layer(
        self,
        layer: Layer,
        pad_rel: float | None = None,
        maybe_empty: bool = True,
    ):
        """This function will be called when a layer is inserted to the canvas."""
        if not self.autoscale_enabled:
            return
        if pad_rel is None:
            pad_rel = 0 if layer._NO_PADDING_NEEDED else 0.025
        xmin, xmax, ymin, ymax, zmin, zmax = layer.bbox_hint()
        if len(self.layers) > 1 or not maybe_empty:
            # NOTE: if there was no layer, so backend may not have xlim/ylim,
            # or they may be set to a default value.
            _xmin, _xmax = self.x.lim
            _ymin, _ymax = self.y.lim
            _dx = (_xmax - _xmin) * pad_rel
            _dy = (_ymax - _ymin) * pad_rel
            xmin = np.min([xmin, _xmin + _dx])
            xmax = np.max([xmax, _xmax - _dx])
            ymin = np.min([ymin, _ymin + _dy])
            ymax = np.max([ymax, _ymax - _dy])

        # this happens when there is <= 1 data
        xmin, xmax = _prep_lims(
            (xmin, xmax),
            self.x.lim,
            pad_rel,
            layer,
            additional=not layer._ATTACH_TO_AXIS
            or getattr(layer, "orient", None) is not Orientation.HORIZONTAL,
        )
        ymin, ymax = _prep_lims(
            (ymin, ymax),
            self.y.lim,
            pad_rel,
            layer,
            additional=not layer._ATTACH_TO_AXIS
            or getattr(layer, "orient", None) is not Orientation.VERTICAL,
        )
        zmin, zmax = _prep_lims((zmin, zmax), self.z.lim, pad_rel, layer)
        self.x.lim = (xmin, xmax)
        self.y.lim = (ymin, ymax)
        self.z.lim = (zmin, zmax)

    def _emit_xlim_changed(self, lim):
        self.x.events.lim.emit(lim)

    def _emit_ylim_changed(self, lim):
        self.y.events.lim.emit(lim)

    def _emit_zlim_changed(self, lim):
        self.z.events.lim.emit(lim)

    @property
    def native(self) -> Any:
        """The native backend object."""
        return self._canvas()._plt_get_native()

    @property
    def aspect_locked(self) -> bool:
        """Whether the aspect ratio is locked."""
        return self._canvas()._plt_get_aspect_locked()

    @aspect_locked.setter
    def aspect_locked(self, value: bool):
        self._canvas()._plt_set_aspect_locked(value)

    def update_axes(
        self,
        *,
        visible: bool | None = None,
        color: ColorType | None = None,
    ):
        """
        Update axes appearance.

        Parameters
        ----------
        visible : bool, optional
            Whether to show the axes.
        color : color-like, optional
            Color of the axes.
        """
        if visible is not None:
            self.x.ticks.visible = self.y.ticks.visible = self.z.ticks.visible = visible
        if color is not None:
            self.x.color = self.y.color = self.z.color = color
            self.x.ticks.color = self.y.ticks.color = self.z.ticks.color = color
            self.x.label.color = self.y.label.color = self.z.label.color = color
        return self

    def update_labels(
        self,
        title: str | None = None,
        x: str | None = None,
        y: str | None = None,
        z: str | None = None,
    ) -> Self:
        """
        Helper function to update the title, x, y and z labels.

        >>> from whitecanvas import new_canvas_3d
        >>> canvas = new_canvas_3d("matplotlib").update_labels("Title", "X", "Y", "Z")
        """
        if title is not None:
            self.title.text = title
            self.title.visible = True
        if x is not None:
            self.x.label.text = x
            self.x.label.visible = True
        if y is not None:
            self.y.label.text = y
            self.y.label.visible = True
        if z is not None:
            self.z.label.text = z
            self.z.label.visible = True
        return self

    def update_font(
        self,
        size: float | None = None,
        color: ColorType | None = None,
        family: str | None = None,
    ) -> Self:
        """
        Update all the fonts, including the title, x/y labels and x/y tick labels.

        Parameters
        ----------
        size : float, optional
            New font size.
        color : color-like, optional
            New font color.
        family : str, optional
            New font family.
        """
        if size is not None:
            self.title.size = size
            self.x.label.size = self.y.label.size = self.z.label.size = size
            self.x.ticks.size = self.y.ticks.size = self.z.ticks.size = size
        if family is not None:
            self.title.family = family
            self.x.label.family = self.y.label.family = self.z.label.family = family
            self.x.ticks.family = self.y.ticks.family = self.z.ticks.family = family
        if color is not None:
            self.title.color = self.x.label.color = self.y.label.color = color
            self.x.ticks.color = self.y.ticks.color = self.z.ticks.color = color
        return self

    @overload
    def add_line(
        self, x: ArrayLike1D, y: ArrayLike1D, z: ArrayLike1D, *,
        name: str | None = None, color: ColorType | None = None,
        width: float | None = None, style: str | LineStyle | None = None,
        alpha: float = 1.0, antialias: bool = True,
    ) -> layer3d.Line3D:  # fmt: skip
        ...

    @overload
    def add_line(
        self, data: Any, *, name: str | None = None, color: ColorType | None = None,
        width: float | None = None, style: str | LineStyle | None = None,
        alpha: float = 1.0, antialias: bool = True,
    ) -> layer3d.Line3D:  # fmt: skip
        ...

    def add_line(
        self,
        *args,
        name=None,
        color=None,
        width=None,
        style=None,
        alpha=1.0,
        antialias=True,
    ) -> layer3d.Line3D:
        xdata, ydata, zdata = normalize_xyz(*args)
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        width = theme._default("line.width", width)
        style = theme._default("line.style", style)
        layer = layer3d.Line3D(
            xdata, ydata, zdata, name=name, color=color, width=width, style=style,
            alpha=alpha, antialias=antialias, backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    @overload
    def add_markers(
        self, data: Any, *, name: str | None = None, symbol: str | None = None,
        size: float | None = None, color: ColorType | None = None, alpha: float = 1.0,
        hatch: str | Hatch | None = None,
    ) -> layer3d.Markers3D:  # fmt: skip
        ...

    @overload
    def add_markers(
        self, x: ArrayLike1D, y: ArrayLike1D, z: ArrayLike1D, *,
        name: str | None = None, symbol: str | None = None, size: float | None = None,
        color: ColorType | None = None, alpha: float = 1.0,
        hatch: str | Hatch | None = None,
    ) -> layer3d.Markers3D:  # fmt: skip
        ...

    def add_markers(
        self,
        *args,
        name=None,
        symbol=None,
        size=None,
        color=None,
        alpha=1.0,
        hatch=None,
    ) -> layer3d.Markers3D:
        """
        Add markers (scatter plot).

        >>> canvas.add_markers(x, y)  # standard usage
        >>> canvas.add_markers(y)  # use 0, 1, ... for the x values

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        symbol : str or Symbol, optional
            Marker symbols. Use the theme default if not specified.
        size : float, optional
            Marker size. Use the theme default if not specified.
        color : color-like, optional
            Color of the marker faces.
        alpha : float, default 1.0
            Alpha channel of the marker faces.
        hatch : str or Hatch, optional
            Pattern of the marker faces. Use the theme default if not specified.

        Returns
        -------
        Markers
            The markers layer.
        """
        xdata, ydata, zdata = normalize_xyz(*args)
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        symbol = theme._default("markers.symbol", symbol)
        size = theme._default("markers.size", size)
        hatch = theme._default("markers.hatch", hatch)
        layer = layer3d.Markers3D(
            xdata, ydata, zdata, name=name, symbol=symbol, size=size, color=color,
            alpha=alpha, hatch=hatch, backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    def add_mesh(
        self,
        verts: NDArray[np.floating],
        faces: NDArray[np.intp],
        *,
        name: str | None = None,
        color: ColorType | None = None,
        hatch: str | Hatch | None = None,
        alpha: float = 1.0,
    ) -> layer3d.Mesh3D:
        """
        Add a mesh layer.

        Parameters
        ----------
        verts : 2D array-like
            Vertices of the mesh.
        faces : 2D array-like
            Faces of the mesh.
        name : str, optional
            Name of the layer.
        color : color-like, optional
            Color of the surface.
        hatch : str or Hatch, optional
            Hatch pattern of the surface faces.
        alpha : float, default 1.0
            Alpha channel of the surface faces.
        """
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        hatch = theme._default("markers.hatch", hatch)
        layer = layer3d.Mesh3D(
            verts, faces, name=name, color=color, hatch=hatch, alpha=alpha,
            backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    @overload
    def add_surface(
        self, z: NDArray[np.floating], *, name: str | None =None,
        color: ColorType | None = None, hatch: str | Hatch | None = None,
        alpha: float = 1.0,
    ) -> layer3d.Mesh3D:  # fmt: skip
        ...

    @overload
    def add_surface(
        self, x: NDArray[np.floating], y: NDArray[np.floating], z: NDArray[np.floating],
        *, name: str | None =None, color: ColorType | None = None,
        hatch: str | Hatch | None = None, alpha: float = 1.0,
    ) -> layer3d.Mesh3D:  # fmt: skip
        ...

    def add_surface(
        self,
        *args,
        name: str | None = None,
        color: ColorType | None = None,
        hatch: str | Hatch | None = None,
        alpha: float = 1.0,
    ) -> layer3d.Mesh3D:
        """
        Add a surface plot defined by z = f(x, y).

        The x, y are usually calculated by `np.meshgrid` or `np.indices`.

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : color-like, optional
            Color of the surface.
        hatch : str or Hatch, optional
            Hatch pattern of the surface faces.
        alpha : float, default 1.0
            Alpha channel of the surface faces.
        """
        if len(args) == 1:
            z = np.asarray(args[0])
            if z.ndim != 2:
                raise ValueError(f"Expected 2D array for z, got shape {z.shape}")
            x, y = np.indices(z.shape)
        elif len(args) == 3:
            x = np.asarray(args[0])
            y = np.asarray(args[1])
            z = np.asarray(args[2])
        else:
            raise TypeError(f"Expected 1 or 3 positional arguments, got {len(args)}")
        nx, ny = x.shape
        verts = np.stack((x.ravel(), y.ravel(), z.ravel()), axis=1)
        # (i, j), (i + 1, j), (i, j + 1)
        # (i + 1, j + 1), (i + 1, j), (i, j + 1)
        corners = (np.arange(nx - 1)[..., None] + np.arange(ny - 1) * nx).ravel()
        face0 = np.stack((corners, corners + 1, corners + nx), axis=1)
        face1 = np.stack((corners + 1, corners + nx + 1, corners + nx), axis=1)
        faces = np.concatenate((face0, face1), axis=0)
        return self.add_mesh(
            verts, faces, name=name, color=color, hatch=hatch, alpha=alpha
        )

    def add_vectors(
        self,
        x: ArrayLike1D,
        y: ArrayLike1D,
        z: ArrayLike1D,
        vx: ArrayLike1D,
        vy: ArrayLike1D,
        vz: ArrayLike1D,
        *,
        name: str | None = None,
        color: ColorType | None = None,
        width: float | None = None,
        style: LineStyle | str | None = None,
        alpha: float = 1.0,
        antialias: bool = True,
    ) -> layer3d.Vectors3D:
        """
        Add a 3D vector field.

        Parameters
        ----------
        x, y, z : array-like
            Base points of the vectors.
        vx, vy, vz : array-like
            Components of the vectors.
        name : str, optional
            Name of the layer.
        color : color-like, optional
            Color of the vectors.
        width : float, optional
            Width of the vectors.
        style : str or LineStyle, optional
            Style of the vectors.
        alpha : float, default 1.0
            Alpha channel of the vectors.
        antialias : bool, default True
            Whether to use antialiasing.
        """
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        width = theme._default("line.width", width)
        style = theme._default("line.style", style)
        layer = layer3d.Vectors3D(
            as_array_1d(x, dtype=np.float32), as_array_1d(y, dtype=np.float32),
            as_array_1d(z, dtype=np.float32), as_array_1d(vx, dtype=np.float32),
            as_array_1d(vy, dtype=np.float32), as_array_1d(vz, dtype=np.float32),
            name=name, color=color, width=width, style=style,
            alpha=alpha, antialias=antialias, backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    def add_layer(self, layer: _L3D) -> _L3D:
        """Add a layer to the canvas."""
        self.layers.append(layer)
        return layer


def _prep_lims(new_lim, current_lim, pad_rel, layer: Layer, additional=False):
    small_diff = 1e-6
    xmin, xmax = new_lim
    if np.isnan(xmax) or np.isnan(xmin):
        xmin, xmax = current_lim
    elif xmax - xmin < small_diff:
        xmin -= 0.05
        xmax += 0.05
    else:
        dx = (xmax - xmin) * pad_rel
        if xmin != 0 or additional:
            xmin -= dx
        xmax += dx
    return xmin, xmax


class Canvas3D(Canvas3DBase):
    _CURRENT_INSTANCE: Canvas3D | None = None

    def __init__(
        self,
        backend: str | None = None,
        *,
        palette: ColormapType | None = None,
    ):
        self._backend = Backend(backend)
        self._backend_object = self._create_backend_object()
        super().__init__(palette=palette)
        self.__class__._CURRENT_INSTANCE = self

    @classmethod
    def from_backend(
        cls,
        obj: protocols.CanvasProtocol,
        *,
        palette: ColormapType | None = None,
        backend: str | None = None,
    ) -> Self:
        """Create a canvas object from a backend object."""
        with patch_dummy_backend() as name:
            # this patch will delay initialization by "_init_canvas" until the backend
            # objects are created.
            self = cls(backend=name, palette=palette)
        self._backend = Backend(backend)
        self._backend_object = obj
        # self._init_canvas()
        return self

    def _create_backend_object(self) -> protocols.CanvasProtocol:
        return self._backend.get_submodule("canvas3d").Canvas3D()

    def _get_backend(self):
        return self._backend

    def _canvas(self) -> protocols.CanvasProtocol:
        return self._backend_object


class _Canvas3DWithGrid(Canvas3DBase):
    def __init__(self, canvas: Canvas3D, grid: CanvasGrid):
        self._main_canvas = canvas
        self._grid = grid
        super().__init__(palette=canvas._color_palette)

    def _get_backend(self) -> Backend:
        """Return the backend."""
        return self._main_canvas._backend

    def _canvas(self):
        return self._main_canvas._backend_object

    @property
    def native(self) -> Any:
        """The native backend object."""
        return self._main_canvas.native

    def show(self, block: bool = False) -> None:
        """Show the canvas using the method defined in the backend."""
        self._grid.show(block=block)

    @property
    def background_color(self) -> NDArray[np.floating]:
        """Background color of the canvas."""
        return self._grid.background_color

    @background_color.setter
    def background_color(self, color):
        self._grid.background_color = color

    @property
    def size(self) -> tuple[float, float]:
        """Size of the canvas"""
        return self._grid.size

    @size.setter
    def size(self, size: tuple[float, float]):
        self._grid.size = size

    def screenshot(self) -> NDArray[np.uint8]:
        """Return a screenshot of the grid."""
        return self._grid.screenshot()

    def _repr_png_(self):
        """Return PNG representation of the widget for QtConsole."""
        return self._grid._repr_png_()

    def _repr_mimebundle_(self, *args: Any, **kwargs: Any) -> dict:
        return self._grid._repr_mimebundle_(*args, **kwargs)

    def _ipython_display_(self, *args: Any, **kwargs: Any) -> Any:
        return self._grid._ipython_display_(*args, **kwargs)

    def _repr_html_(self, *args: Any, **kwargs: Any) -> str:
        return self._grid._repr_html_(*args, **kwargs)

    def to_html(self, file: str | None = None) -> str:
        """Return HTML representation of the canvas."""
        return self._grid.to_html(file=file)


class SingleCanvas3D(_Canvas3DWithGrid):
    """
    A canvas without other subplots.

    This class is the simplest form of canvas. In `matplotlib` terms, it is a figure
    with a single axes.
    """

    def __init__(self, grid: CanvasGrid):
        if grid.shape != (1, 1):
            raise ValueError(f"Grid shape must be (1, 1), got {grid.shape}")
        self._grid = grid
        _it = grid._iter_canvas()
        _, canvas = next(_it)
        if next(_it, None) is not None:
            raise ValueError("Grid must have only one canvas")
        self._main_canvas = canvas
        super().__init__(canvas, grid)

        # NOTE: events, dims etc are not shared between the main canvas and the
        # SingleCanvas instance. To avoid confusion, the first and the only canvas
        # should be replaces with the SingleCanvas instance.
        # self.mouse = grid[0, 0].mouse
        grid._canvas_array[0, 0] = self
        # self.events.drawn.connect(
        #     self._main_canvas.events.drawn.emit, unique=True, max_args=None
        # )



from __future__ import annotations

from typing import Callable

import numpy as np
from mpl_toolkits.mplot3d import Axes3D, art3d

from whitecanvas.backend.matplotlib._base import MplLayer, MplMouseEventsMixin
from whitecanvas.backend.matplotlib._labels import MplAxis, MplLabel, MplTicks, Title
from whitecanvas.backend.matplotlib.components3d.vectors3d import Vectors3D


class Canvas3D:
    def __init__(self, ax: Axes3D):
        self._axes = ax
        self._xaxis = MplAxis(self, "x")
        self._yaxis = MplAxis(self, "y")
        self._zaxis = MplAxis(self, "z")
        self._title = Title(self)
        self._xlabel = MplLabel(self, "x")
        self._ylabel = MplLabel(self, "y")
        self._zlabel = MplLabel(self, "z")
        self._xticks = MplTicks(self, "x")
        self._yticks = MplTicks(self, "y")
        self._zticks = MplTicks(self, "z")

    def _plt_add_layer(self, layer):
        if isinstance(layer, art3d.Line3D):
            self._axes.add_line(layer)
        elif isinstance(layer, (art3d.Path3DCollection, Vectors3D)):
            self._axes.add_collection3d(layer)
        elif isinstance(layer, art3d.Poly3DCollection):
            self._axes.add_collection(layer)
        else:
            raise NotImplementedError(f"{layer}")
        if hasattr(layer, "post_add"):
            layer.post_add(self)

    def _plt_get_native(self):
        return self._axes

    def _plt_get_title(self):
        return self._title

    def _plt_get_xaxis(self):
        return self._xaxis

    def _plt_get_yaxis(self):
        return self._yaxis

    def _plt_get_zaxis(self):
        return self._zaxis

    def _plt_get_xlabel(self):
        return self._xlabel

    def _plt_get_ylabel(self):
        return self._ylabel

    def _plt_get_zlabel(self):
        return self._zlabel

    def _plt_get_xticks(self):
        return self._xticks

    def _plt_get_yticks(self):
        return self._yticks

    def _plt_get_zticks(self):
        return self._zticks

    def _plt_reorder_layers(self, layers: list[MplLayer]):
        for i, layer in enumerate(layers):
            layer._plt_set_zorder(i)

    def _plt_get_aspect_locked(self) -> bool:
        out = self._axes.get_aspect()
        return out == "equal"

    def _plt_set_aspect_locked(self, locked: bool):
        if locked:
            self._axes.set_aspect("equal")
        else:
            self._axes.set_aspect("auto")

    def _plt_connect_xlim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        """Connect callback to x-limits changed event"""
        self._axes.callbacks.connect("xlim_changed", lambda ax: callback(ax.get_xlim()))

    def _plt_connect_ylim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        """Connect callback to y-limits changed event"""
        self._axes.callbacks.connect("ylim_changed", lambda ax: callback(ax.get_ylim()))

    def _plt_connect_zlim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        """Connect callback to z-limits changed event"""
        self._axes.callbacks.connect("zlim_changed", lambda ax: callback(ax.get_zlim()))

    def _plt_draw(self):
        if fig := self._axes.get_figure():
            fig.canvas.draw_idle()



from whitecanvas.backend.bokeh._base import to_html
from whitecanvas.backend.bokeh.band import Band
from whitecanvas.backend.bokeh.bars import Bars
from whitecanvas.backend.bokeh.canvas import Canvas, CanvasGrid
from whitecanvas.backend.bokeh.image import Image
from whitecanvas.backend.bokeh.line import MonoLine, MultiLine
from whitecanvas.backend.bokeh.markers import Markers
from whitecanvas.backend.bokeh.text import Texts



from whitecanvas.layers.group._collections import (
    LayerCollectionBase,
    LayerTuple,
    MainAndOtherLayers,
)
from whitecanvas.layers.group.band_collection import BandCollection, ViolinPlot
from whitecanvas.layers.group.boxplot import BoxPlot
from whitecanvas.layers.group.colorbar import Colorbar
from whitecanvas.layers.group.graph import Graph
from whitecanvas.layers.group.labeled import (
    LabeledBars,
    LabeledImage,
    LabeledLine,
    LabeledMarkers,
    LabeledPlot,
)
from whitecanvas.layers.group.line_band import LineBand
from whitecanvas.layers.group.line_collection import LineCollection
from whitecanvas.layers.group.line_fill import Area, Histogram, Kde, LineFillBase
from whitecanvas.layers.group.line_markers import Plot
from whitecanvas.layers.group.marker_collection import MarkerCollection
from whitecanvas.layers.group.stemplot import StemPlot
from whitecanvas.layers.group.textinfo import BracketText, Panel

__all__ = [
    "Area",
    "Plot",
    "LineBand",
    "BandCollection",
    "BracketText",
    "Colorbar",
    "Panel",
    "MainAndOtherLayers",
    "LabeledLine",
    "LabeledMarkers",
    "LabeledBars",
    "LabeledPlot",
    "LabeledImage",
    "LineFillBase",
    "ViolinPlot",
    "MarkerCollection",
    "LineCollection",
    "BoxPlot",
    "Graph",
    "StemPlot",
    "LayerTuple",
    "Histogram",
    "Kde",
    "LayerCollectionBase",
]



"""Layer with a dataframe bound to it."""

from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Generic, Literal, Sequence, TypeVar

import numpy as np
from cmap import Color

from whitecanvas import theme
from whitecanvas.backend import Backend
from whitecanvas.layers import Layer, _legend, _mixin
from whitecanvas.layers import group as _lg
from whitecanvas.layers.tabular import _jitter, _shared
from whitecanvas.layers.tabular import _plans as _p
from whitecanvas.layers.tabular._df_compat import DataFrameWrapper, parse
from whitecanvas.types import (
    ColormapType,
    ColorType,
    Hatch,
    LineStyle,
    Orientation,
    Symbol,
    _Void,
)
from whitecanvas.utils.type_check import is_real_number

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.canvas.dataframe._base import CatIterator
    from whitecanvas.layers.tabular import DFMarkerGroups, DFRugGroups

    _FE = _mixin.AbstractFaceEdgeMixin[_mixin.FaceNamespace, _mixin.EdgeNamespace]

_DF = TypeVar("_DF")
_L = TypeVar("_L", bound=Layer)
_void = _Void()


def _norm_color_hatch(
    color,
    hatch,
    df: DataFrameWrapper[_DF],
) -> tuple[_p.ColorPlan, _p.HatchPlan]:
    color_cov = _shared.ColumnOrValue(color, df)
    if color_cov.is_column:
        color_by = _p.ColorPlan.from_palette(color_cov.columns)
    elif color_cov.value is not None:
        color_by = _p.ColorPlan.from_const(Color(color_cov.value))
    else:
        color_by = _p.ColorPlan.default()
    hatch_cov = _shared.ColumnOrValue(hatch, df)
    if hatch_cov.is_column:
        hatch_by = _p.HatchPlan.new(hatch_cov.columns)
    elif hatch_cov.value is not None:
        hatch_by = _p.HatchPlan.from_const(Hatch(hatch_cov.value))
    else:
        hatch_by = _p.HatchPlan.default()
    return color_by, hatch_by


class _BoxLikeMixin:
    _source: DataFrameWrapper[_DF]

    def __init__(
        self,
        categories: list[tuple],
        splitby: tuple[str, ...],
        color_by: _p.ColorPlan,
        hatch_by: _p.HatchPlan,
    ):
        self._splitby = splitby
        self._categories = categories
        self._color_by = color_by
        self._hatch_by = hatch_by
        self._get_base().face.color = color_by.generate(self._categories, self._splitby)
        self._get_base().face.hatch = hatch_by.generate(self._categories, self._splitby)
        if not hasattr(self, "with_hover_template"):
            return
        self.with_hover_template("\n".join(f"{k}: {{{k}!r}}" for k in self._splitby))

    def _get_base(self) -> _FE:
        """Just for typing."""
        return self._base_layer

    def _normalize_by_arg(self, by, default: tuple[str, ...]) -> tuple[str, ...]:
        if by is None:
            by = default
        elif isinstance(by, str):
            if by not in self._splitby:
                raise ValueError(
                    f"Cannot color by {by!r} as the plot is not split by this column. "
                    f"Valid columns are: {self._splitby!r}."
                )
            by = (by,)
        else:
            for b in by:
                if not isinstance(b, str):
                    raise TypeError("`by` must be a str or sequence of str.")
                if b not in self._splitby:
                    raise ValueError(
                        f"Cannot color by {by!r} as the plot is not split by this "
                        f"column. Valid columns are: {self._splitby!r}."
                    )
        return by

    def update_color_palette(
        self,
        palette: ColormapType | None = None,
        *,
        alpha: float | None = None,
        cycle_by: str | Sequence[str] | None = None,
    ) -> Self:
        """
        Update the colors by a color palette.

        Parameters
        ----------
        palette : colormap type
            Color palette used to generate colors for each category. A color palette
            can be a list of colors or any types that can be converted into a
            `cmap.Colormap` object.
        alpha : float, optional
            Additional alpha value that will be applied to the palette colors.
        cycle_by : str or sequence of str, optional
            If given, colors will be cycled on this column name(s).
        """
        by = self._normalize_by_arg(cycle_by, self._color_by.by)
        color_by = _p.ColorPlan.from_palette(by, palette=palette)
        colors = color_by.generate(self._categories, self._splitby)
        color_arr = np.stack([c.rgba for c in colors], dtype=np.float32)
        if alpha is not None:
            if is_real_number(alpha) and 0 <= alpha <= 1:
                color_arr[:, 3] = alpha
            else:
                raise TypeError(
                    f"`alpha` must be a scalar value between 0 and 1, got {alpha!r}."
                )
        self._get_base().face.color = color_arr
        self._color_by = color_by
        return self

    def update_hatch_palette(
        self,
        palette: Sequence[str | Hatch],
        *,
        cycle_by: str | Sequence[str] | None = None,
    ) -> Self:
        """
        Update the hatch patterns by a list of hatch values.

        Parameters
        ----------
        palette : sequence of str or Hatch
            Hatch palette used to generate colors for each category.
        """
        by = self._normalize_by_arg(cycle_by, self._hatch_by.by)
        hatch_by = _p.HatchPlan.new(by, values=palette)
        self._get_base().face.hatch = hatch_by.generate(self._categories, self._splitby)
        self._hatch_by = hatch_by
        return self

    def update_const(
        self,
        *,
        color: ColorType | _Void = _void,
        hatch: str | Hatch | _Void = _void,
        alpha: float | _Void = _void,
    ) -> Self:
        """
        Update the plot features to the constant values.

        Parameters
        ----------
        color : color-type, optional
            Constant colors used for the plot.
        hatch : str or Hatch, optional
            Constant hatch used for the plot.
        """
        cat = self._categories
        if color is not _void:
            color_by = _p.ColorPlan.from_const(Color(color))
            self._get_base().face.color = color_by.generate(cat, self._splitby)
            self._color_by = color_by
        if hatch is not _void:
            hatch_by = _p.HatchPlan.from_const(Hatch(hatch))
            self._get_base().face.hatch = hatch_by.generate(cat, self._splitby)
            self._hatch_by = hatch_by
        if alpha is not _void:
            self._get_base().face.alpha = alpha
            self._get_base().edge.alpha = alpha
        return self

    def with_face(
        self,
        *,
        color: ColorType | None = None,
        hatch: str | Hatch | None = None,
        alpha: float = 1.0,
    ) -> Self:
        """Add face to the plot with given settings."""
        if color is not None:
            self._get_base().with_face(color=color, hatch=hatch, alpha=alpha)
        return self

    def with_edge(
        self,
        *,
        color: ColorType | None = None,
        width: float = 1.0,
        style: str | LineStyle = LineStyle.SOLID,
        alpha: float = 1.0,
    ) -> Self:
        """Add edge to the plot with given settings."""
        self._get_base().with_edge(color=color, width=width, style=style, alpha=alpha)
        return self

    def _prep_legend_info(
        self,
    ) -> tuple[list[tuple[str, ColorType]], list[tuple[str, Hatch]]]:
        df = self._simple_dataframe()
        color_entries = self._color_by.to_entries(df)
        hatch_entries = self._hatch_by.to_entries(df)
        return color_entries, hatch_entries

    def _as_legend_item(self) -> _legend.LegendItemCollection:
        colors, hatches = self._prep_legend_info()
        items = []
        color_default = theme.get_theme().background_color
        edge_info = self._get_base().edge._as_legend_info()
        if self._color_by.is_const():
            color_default = Color(self._color_by.values[0])
        else:
            items.append((", ".join(self._color_by.by), _legend.TitleItem()))
            for label, color in colors:
                item = _legend.BarLegendItem(_legend.FaceInfo(color), edge_info)
                items.append((label, item))
        if self._hatch_by.is_not_const():
            items.append((", ".join(self._hatch_by.by), _legend.TitleItem()))
            for label, hatch in hatches:
                item = _legend.BarLegendItem(
                    _legend.FaceInfo(color_default, hatch), edge_info
                )
                items.append((label, item))
        return _legend.LegendItemCollection(items)

    def _simple_dataframe(self) -> DataFrameWrapper[dict]:
        return _shared.list_to_df(self._categories, self._splitby)


class DFViolinPlot(
    _shared.DataFrameLayerWrapper[_lg.ViolinPlot, _DF],
    _BoxLikeMixin,
    Generic[_DF],
):
    def __init__(
        self,
        cat: CatIterator[_DF],
        value: str,
        color: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
        dodge: str | tuple[str, ...] | bool | None = None,
        name: str | None = None,
        orient: Orientation = Orientation.VERTICAL,
        extent: float = 0.8,
        shape: Literal["both", "left", "right"] = "both",
        backend: str | Backend | None = None,
    ):
        _splitby, dodge = _shared.norm_dodge(
            cat.df, cat.offsets, color, hatch, dodge=dodge
        )  # fmt: skip
        x, arr, categories = cat.prep_arrays(_splitby, value, dodge=dodge)
        _extent = cat.zoom_factor(dodge=dodge) * extent
        color_by, hatch_by = _norm_color_hatch(color, hatch, cat.df)
        base = _lg.ViolinPlot.from_arrays(
            x, arr, name=name, orient=orient, shape=shape, extent=_extent,
            backend=backend,
        )  # fmt: skip
        super().__init__(base, cat.df)
        _BoxLikeMixin.__init__(self, categories, _splitby, color_by, hatch_by)
        self._offsets = cat.offsets
        self._value = value
        self._dodge = dodge
        self._numeric = cat._numeric
        self.with_hover_template("\n".join(f"{k}: {{{k}!r}}" for k in self._splitby))

    @property
    def orient(self) -> Orientation:
        """Orientation of the violins."""
        return self._base_layer.orient

    def move(self, shift: float = 0.0, autoscale: bool = True) -> Self:
        """Move the layer by the given shift."""
        for layer in self._base_layer:
            _old = layer.data
            layer.set_data(edge_low=_old.y0 + shift, edge_high=_old.y1 + shift)
        if autoscale and (canvas := self._canvas_ref()):
            canvas._autoscale_for_layer(self, pad_rel=0.025)
        return self

    def with_hover_text(self, text: str | list[str]) -> Self:
        """Set the hover tooltip text for the layer."""
        self.base.with_hover_text(text)
        return self

    def with_hover_template(self, template: str) -> Self:
        """Set the hover tooltip template for the layer."""
        extra = {}
        for i, key in enumerate(self._splitby):
            extra[key] = [row[i] for row in self._categories]
        self.base.with_hover_template(template, extra=extra)
        return self

    def with_rug(
        self,
        *,
        width: float = 1.0,
        color: ColorType | None = None,
    ) -> _lg.MainAndOtherLayers[Self, DFRugGroups[_DF]]:
        """Overlay rug plot on the violins and return the violin layer."""
        from whitecanvas.layers.tabular import DFRugGroups

        canvas = self._canvas_ref()
        if canvas is None:
            raise ValueError("No canvas to add the rug plot.")
        _extent = self.base.extent
        if color is not None:
            colors = Color(color)
        elif self._is_edge_only():
            colors = self._color_by.by
        else:
            colors = Color("#1F1F1F")
        jitter = _jitter.CategoricalJitter(
            self._splitby,
            self._make_cat_iterator().prep_position_map(self._splitby, self._dodge),
        )
        if self.base._shape == "both":
            align = "center"
        elif self.base._shape == "left":
            align = "high"
        else:
            align = "low"
        rug = DFRugGroups.from_table(
            self._source, jitter, self._value, color=colors, width=width,
            extent=_extent, backend=self.base._backend_name,
        ).scale_by_density(align=align)  # fmt: skip
        return _combine_main_and_others(self, rug)

    def with_box(
        self,
        *,
        color: ColorType | None = None,
        median_color: ColorType = "white",
        width: float | None = None,
        extent: float = 0.1,
        capsize: float = 0.0,
    ) -> _lg.MainAndOtherLayers[Self, DFBoxPlot[_DF]]:
        """
        Overlay box plot on the violins and return the violin layer.

        Following the convension of many statistical software, the box plot is colored
        by black if the violin faces are colored, and colored by the edge color
        otherwise. The median line is colored by the given median color.

        Parameters
        ----------
        color : color-type, optional
            Color of the box plot. If not given, it will be colored by "#1F1F1F" if
            the violin faces are colored, and by the edge color of the violin plot
            otherwise.
        median_color : color-type, optional
            Color of the median line of the box plot.
        width : float, optional
            Width of the whiskers of the boxplot. Use violin edge width if not given.
        extent : float, optional
            Relative width of the boxes.
        capsize : float, optional
            Relative size of the caps of the whiskers.
        """

        canvas = self._canvas_ref()
        if canvas is None:
            raise ValueError("No canvas to add the box plot.")
        if color is not None:
            colors = Color(color)
        else:
            if np.all(self.base.edge.width > 0) and np.all(self.base.edge.alpha > 0):
                colors = self.base.edge.color
            else:
                colors = Color("#1F1F1F")
        if width is None:
            width = self.base.edge.width.mean()
        box = DFBoxPlot(
            self._make_cat_iterator(), self._value, name=f"boxplot-of-{self.name}",
            color=None, hatch=Hatch.SOLID, dodge=self._dodge, width=width,
            orient=self.orient, capsize=capsize, extent=extent,
            backend=canvas._get_backend(),
        )  # fmt: skip
        box.base.boxes.face.color = colors
        box.base.edge.color = colors
        box.base.medians.color = Color(median_color)
        return _combine_main_and_others(self, box)

    def with_outliers(
        self,
        *,
        color: ColorType | None = None,
        symbol: str | Symbol = Symbol.CIRCLE,
        size: float | None = None,
        ratio: float = 1.5,
        extent: float = 0.1,
        seed: int | None = 0,
    ) -> _lg.MainAndOtherLayers[Self, DFMarkerGroups[_DF]]:
        """
        Overlay outliers on the box plot and return the box plot layer.

        Parameters
        ----------
        color : color-type, optional
            Color of the outliers. To make sure the outliers are easily visible, face
            color will always be transparent. If a constant color is given, all the
            edges will be colored by the same color. By default, the edge colors are
            the same as the edge colors of the box plot.
        symbol : str or Symbol, optional
            Symbol of the outlier markers.
        size : float, optional
            Size of the outlier markers. If not given, it will be set to the theme
            default.
        ratio : float, optional
            Ratio of the interquartile range (IQR) to determine the outliers. Data
            points outside of the range [Q1 - ratio * IQR, Q3 + ratio * IQR] will be
            considered as outliers.
        extent : float, optional
            Relative width of the jitter range (same effect as the `extent` argument of
            the `add_stripplot` method).
        seed : int, optional
            Random seed for the jitter (same effect as the `seed` argument of the
            `add_stripplot` method).
        """
        from whitecanvas.layers.tabular import DFMarkerGroups

        canvas = self._canvas_ref()
        size = theme._default("markers.size", size)
        if canvas is None:
            raise ValueError("No canvas to add the outliers.")

        is_edge_only = self._is_edge_only()

        # category iterator is used to calculate positions and indices
        _cat_self = self._make_cat_iterator()
        _pos_map = _cat_self.prep_position_map(self._splitby, self._dodge)
        _extent = _cat_self.zoom_factor(self._dodge) * extent

        # calculate outliers and update the separators
        df_outliers = {c: [] for c in (*self._splitby, self._value)}
        colors = []
        for idx_cat, (sl, sub) in enumerate(self._source.group_by(self._splitby)):
            arr = sub[self._value]
            q1, q3 = np.quantile(arr, [0.25, 0.75])
            iqr = q3 - q1  # interquartile range
            low = q1 - ratio * iqr  # lower bound of inliers
            high = q3 + ratio * iqr  # upper bound of inliers
            outliers = arr[(arr < low) | (arr > high)]
            for _cat, _s in zip(sl, self._splitby):
                df_outliers[_s].extend([_cat] * outliers.size)
            df_outliers[self._value].extend(outliers)
            if is_edge_only:
                _this_color = self.base.edge.color[idx_cat]
            else:
                _this_color = self.base.face.color[idx_cat]
            colors.extend([_this_color] * outliers.size)

        df_outliers = parse(df_outliers)
        xj = _jitter.UniformJitter(self._splitby, _pos_map, extent=_extent, seed=seed)
        yj = _jitter.IdentityJitter(self._value).check(df_outliers)
        new = DFMarkerGroups(
            df_outliers, xj, yj, name=f"outliers-of-{self.name}", color=Color("black"),
            orient=self.orient, symbol=symbol, size=size, backend=canvas._get_backend(),
        )  # fmt: skip
        if color is None:
            if is_edge_only:  # edge only
                new._apply_color(np.stack(colors, axis=0, dtype=np.float32))
            new.as_edge_only(width=self.base.edge.width.mean())
        return _combine_main_and_others(self, new)

    def with_strip(
        self,
        *,
        color: ColorType | None = None,
        symbol: str | Symbol = Symbol.CIRCLE,
        size: str | None = None,
        extent: float = 0.2,
        seed: int | None = 0,
    ) -> _lg.MainAndOtherLayers[Self, DFMarkerGroups[_DF]]:
        """
        Overlay strip plot on the violins.

        Parameters
        ----------
        color : color-type, optional
            Color of the strip plot. If not given, it will be colored by the violin
            face color.
        symbol : str or Symbol, optional
            Symbol of the strip plot markers.
        size : float, optional
            Size of the strip plot markers. If not given, it will be set to the theme
            default.
        extent : float, optional
            Relative width of the jitter range.
        seed : int, optional
            Random seed for the jitter.
        """
        from whitecanvas.layers.tabular import DFMarkerGroups

        canvas = self._canvas_ref()
        size = theme._default("markers.size", size)
        if canvas is None:
            raise ValueError("No canvas to add the outliers.")

        if color is None:
            color = self._color_by.by
        else:
            color = Color(color)

        # category iterator is used to calculate positions and indices
        _cat_self = self._make_cat_iterator()
        _pos_map = _cat_self.prep_position_map(self._splitby, self._dodge)
        _extent = _cat_self.zoom_factor(self._dodge) * extent
        df = self._source
        xj = _jitter.UniformJitter(self._splitby, _pos_map, extent=_extent, seed=seed)
        yj = _jitter.IdentityJitter(self._value).check(df)
        new = DFMarkerGroups(
            df, xj, yj, name=f"outliers-of-{self.name}", color=color,
            orient=self.orient, symbol=symbol, size=size, backend=canvas._get_backend(),
        )  # fmt: skip
        if self._is_edge_only():
            new.as_edge_only(width=self.base.edge.width.mean())
        return _combine_main_and_others(self, new)

    def with_swarm(
        self,
        *,
        color: ColorType | None = None,
        symbol: str | Symbol = Symbol.CIRCLE,
        size: str | None = None,
        extent: float = 0.8,
        sort: bool = False,
    ) -> _lg.MainAndOtherLayers[Self, DFMarkerGroups[_DF]]:
        """
        Overlay swarm plot on the violins.

        Parameters
        ----------
        color : color-type, optional
            Color of the strip plot. If not given, it will be colored by the violin
            face color.
        symbol : str or Symbol, optional
            Symbol of the strip plot markers.
        size : float, optional
            Size of the strip plot markers. If not given, it will be set to the theme
            default.
        extent : float, optional
            Relative width of the jitter range.
        sort : bool, default False
            If True, the markers will be sorted by the value.
        """
        from whitecanvas.layers.tabular import DFMarkerGroups

        canvas = self._canvas_ref()
        size = theme._default("markers.size", size)
        if canvas is None:
            raise ValueError("No canvas to add the outliers.")

        if color is None:
            color = self._color_by.by
        else:
            color = Color(color)

        # category iterator is used to calculate positions and indices
        _cat_self = self._make_cat_iterator()
        _pos_map = _cat_self.prep_position_map(self._splitby, self._dodge)
        _extent = _cat_self.zoom_factor(self._dodge) * extent
        df = self._source

        if sort:
            df = df.sort(self._value)
        lims = df[self._value].min(), df[self._value].max()
        xj = _jitter.SwarmJitter(
            self._splitby, _pos_map, self._value, lims, extent=_extent
        )
        yj = _jitter.IdentityJitter(self._value).check(df)
        new = DFMarkerGroups(
            df, xj, yj, name=f"outliers-of-{self.name}", color=color,
            orient=self.orient, symbol=symbol, size=size, backend=canvas._get_backend(),
        )  # fmt: skip
        if self._is_edge_only():
            new.as_edge_only(width=self.base.edge.width.mean())
        return _combine_main_and_others(self, new)

    def as_edge_only(
        self,
        width: float = 3.0,
        style: str | LineStyle = LineStyle.SOLID,
    ) -> Self:
        """
        Replace the violin edge color with the face color and delete the face color.

        Parameters
        ----------
        width : float, optional
            Width of the edge.
        style : str or LineStyle, optional
            Style of the edge.
        """
        self.base.with_edge(color=self.base.face.color, width=width, style=style)
        self.base.face.update(alpha=0.0)
        return self

    def _as_legend_item(self) -> _legend.LegendItemCollection:
        return _BoxLikeMixin._as_legend_item(self)

    def _make_cat_iterator(self) -> CatIterator[_DF]:
        from whitecanvas.canvas.dataframe._base import CatIterator

        return CatIterator(self._source, offsets=self._offsets, numeric=self._numeric)

    def _is_edge_only(self) -> bool:
        return np.all(self.base.face.alpha < 1e-6)


class DFBoxPlot(
    _shared.DataFrameLayerWrapper[_lg.BoxPlot, _DF], _BoxLikeMixin, Generic[_DF]
):
    def __init__(
        self,
        cat: CatIterator[_DF],
        value: str,
        color: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
        width: float = 1.0,
        dodge: str | tuple[str, ...] | bool | None = None,
        name: str | None = None,
        orient: Orientation = Orientation.VERTICAL,
        extent: float = 0.8,
        capsize: float = 0.1,
        backend: str | Backend | None = None,
    ):
        _splitby, dodge = _shared.norm_dodge(
            cat.df, cat.offsets, color, hatch, dodge=dodge,
        )  # fmt: skip
        x, arr, categories = cat.prep_arrays(_splitby, value, dodge=dodge)
        _extent = cat.zoom_factor(dodge=dodge) * extent
        _capsize = cat.zoom_factor(dodge=dodge) * capsize
        color_by, hatch_by = _norm_color_hatch(color, hatch, cat.df)
        base = _lg.BoxPlot.from_arrays(
            x, arr, name=name, orient=orient, capsize=_capsize, extent=_extent,
            backend=backend,
        )  # fmt: skip
        base.edge.width = width
        super().__init__(base, cat.df)
        _BoxLikeMixin.__init__(self, categories, _splitby, color_by, hatch_by)
        self._offsets = cat.offsets
        self._value = value
        self._dodge = dodge
        self._numeric = cat._numeric

    @property
    def orient(self) -> Orientation:
        """Orientation of the violins."""
        return self._base_layer.orient

    def move(self, shift: float = 0.0, autoscale: bool = True) -> Self:
        """Move the layer by the given shift."""
        self._base_layer.move(shift, autoscale=autoscale)
        return self

    def with_hover_text(self, text: str | list[str]) -> Self:
        """Set the hover tooltip text for the layer."""
        self.base.boxes.with_hover_text(text)
        return self

    def with_hover_template(self, template: str) -> Self:
        """Set the hover tooltip template for the layer."""
        extra = {}
        for i, key in enumerate(self._splitby):
            extra[key] = [row[i] for row in self._categories]
        self.base.boxes.with_hover_template(template, extra=extra)
        return self

    def with_outliers(
        self,
        *,
        color: ColorType | None = None,
        symbol: str | Symbol = Symbol.CIRCLE,
        size: float | None = None,
        ratio: float = 1.5,
        extent: float = 0.1,
        seed: int | None = 0,
        update_whiskers: bool = True,
    ) -> _lg.MainAndOtherLayers[Self, DFMarkerGroups[_DF]]:
        """
        Overlay outliers on the box plot.

        Parameters
        ----------
        color : color-type, optional
            Color of the outliers. To make sure the outliers are easily visible, face
            color will always be transparent. If a constant color is given, all the
            edges will be colored by the same color. By default, the edge colors are
            the same as the edge colors of the box plot.
        symbol : str or Symbol, optional
            Symbol of the outlier markers.
        size : float, optional
            Size of the outlier markers. If not given, it will be set to the theme
            default.
        ratio : float, optional
            Ratio of the interquartile range (IQR) to determine the outliers. Data
            points outside of the range [Q1 - ratio * IQR, Q3 + ratio * IQR] will be
            considered as outliers.
        extent : float, optional
            Relative width of the jitter range (same effect as the `extent` argument of
            the `add_stripplot` method).
        seed : int, optional
            Random seed for the jitter (same effect as the `seed` argument of the
            `add_stripplot` method).
        update_whiskers : bool, default True
            If True, the whiskers of the box plot will be updated to exclude the
            outliers.
        """
        from whitecanvas.canvas.dataframe._base import CatIterator
        from whitecanvas.layers.tabular import DFMarkerGroups

        canvas = self._canvas_ref()
        size = theme._default("markers.size", size)
        if canvas is None:
            raise ValueError("No canvas to add the outliers.")

        is_edge_only = np.all(self.base.boxes.face.alpha < 1e-6)

        # category iterator is used to calculate positions and indices
        _cat_self = CatIterator(self._source, self._offsets, self._numeric)
        _pos_map = _cat_self.prep_position_map(self._splitby, self._dodge)
        _extent = _cat_self.zoom_factor(self._dodge) * extent

        # calculate outliers and update the separators
        df_outliers = {c: [] for c in (*self._splitby, self._value)}
        agg_values = self.base._get_sep_values()  # for updating whiskers
        colors = []
        for idx_cat, (sl, sub) in enumerate(self._source.group_by(self._splitby)):
            arr = sub[self._value]
            q1, q3 = np.quantile(arr, [0.25, 0.75])
            iqr = q3 - q1  # interquartile range
            low = q1 - ratio * iqr  # lower bound of inliers
            high = q3 + ratio * iqr  # upper bound of inliers
            is_inlier = (low <= arr) & (arr <= high)
            inliers = arr[is_inlier]
            agg_values[0, idx_cat] = inliers.min()
            agg_values[4, idx_cat] = inliers.max()
            outliers = arr[~is_inlier]
            for _cat, _s in zip(sl, self._splitby):
                df_outliers[_s].extend([_cat] * outliers.size)
            df_outliers[self._value].extend(outliers)
            if is_edge_only:
                _this_color = self.base.edge.color[idx_cat]
            else:
                _this_color = self.base.face.color[idx_cat]
            colors.extend([_this_color] * outliers.size)

        df_outliers = parse(df_outliers)
        xj = _jitter.UniformJitter(self._splitby, _pos_map, extent=_extent, seed=seed)
        yj = _jitter.IdentityJitter(self._value).check(df_outliers)
        new = DFMarkerGroups(
            df_outliers, xj, yj, name=f"outliers-of-{self.name}", color=Color("black"),
            orient=self.orient, symbol=symbol, size=size, backend=canvas._get_backend(),
        )  # fmt: skip
        if color is None:
            if is_edge_only:  # edge only
                new._apply_color(np.stack(colors, axis=0, dtype=np.float32))
            new.as_edge_only(width=self.base.edge.width.mean())
        if update_whiskers:
            self.base._update_data(agg_values)
        return _combine_main_and_others(self, new)

    def as_edge_only(
        self,
        width: float = 3.0,
        style: str | LineStyle = LineStyle.SOLID,
    ) -> Self:
        """
        Replace the violin edge color with the face color and delete the face color.

        Parameters
        ----------
        width : float, optional
            Width of the edge.
        style : str or LineStyle, optional
            Style of the edge.
        """
        self.base.with_edge(color=self.base.face.color, width=width, style=style)
        self.base.face.update(alpha=0.0)
        return self

    def _as_legend_item(self) -> _legend.LegendItemCollection:
        return _BoxLikeMixin._as_legend_item(self)


class _EstimatorMixin(_BoxLikeMixin):
    orient: Orientation

    def est_by_mean(self) -> Self:
        """Set estimator to mean."""

        def est_func(x):
            return np.mean(x)

        return self._update_estimate(est_func)

    def est_by_median(self) -> Self:
        """Set estimator to median."""

        def est_func(x):
            return np.median(x)

        return self._update_estimate(est_func)

    def err_by_sd(self, scale: float = 1.0, *, ddof: int = 1) -> Self:
        """Set error to standard deviation."""

        def err_func(x):
            _mean = np.mean(x)
            _sd = np.std(x, ddof=ddof) * scale
            return _mean - _sd, _mean + _sd

        return self._update_error(err_func)

    def err_by_se(self, scale: float = 1.0, *, ddof: int = 1) -> Self:
        """Set error to standard error."""

        def err_func(x):
            _mean = np.mean(x)
            _er = np.std(x, ddof=ddof) / np.sqrt(len(x)) * scale
            return _mean - _er, _mean + _er

        return self._update_error(err_func)

    def err_by_quantile(self, low: float = 0.25, high: float | None = None) -> Self:
        """Set error to quantile."""
        if low < 0 or low > 1:
            raise ValueError(f"Quantile must be between 0 and 1, got {low}")
        if high is None:
            high = 1 - low
        elif high < 0 or high > 1:
            raise ValueError(f"Quantile must be between 0 and 1, got {high}")

        def err_func(x):
            _qnt = np.quantile(x, [low, high])
            return _qnt[0], _qnt[1]

        return self._update_error(err_func)

    def _update_estimate(self, est_func: Callable[[np.ndarray], float]) -> Self:
        arrays = self._get_arrays()
        est = [est_func(arr) for arr in arrays]
        self._set_estimation_values(est)
        return self

    def _update_error(
        self,
        err_func: Callable[[np.ndarray], tuple[float, float]],
    ) -> Self:
        arrays = self._get_arrays()
        err_low = []
        err_high = []
        for arr in arrays:
            low, high = err_func(arr)
            err_low.append(low)
            err_high.append(high)
        self._set_error_values(err_low, err_high)
        return self


class DFPointPlot(
    _shared.DataFrameLayerWrapper[_lg.LabeledPlot, _DF], _EstimatorMixin, Generic[_DF]
):
    def __init__(
        self,
        cat: CatIterator[_DF],
        value: str,
        color: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
        dodge: str | tuple[str, ...] | bool | None = None,
        name: str | None = None,
        orient: Orientation = Orientation.VERTICAL,
        capsize: float = 0.1,
        backend: str | Backend | None = None,
    ):
        _splitby, dodge = _shared.norm_dodge(
            cat.df, cat.offsets, color, hatch, dodge=dodge,
        )  # fmt: skip
        x, arr, categories = cat.prep_arrays(_splitby, value, dodge=dodge)
        _capsize = cat.zoom_factor(dodge=dodge) * capsize
        color_by, hatch_by = _norm_color_hatch(color, hatch, cat.df)
        base = _lg.LabeledPlot.from_arrays(
            x, arr, name=name, orient=orient, capsize=_capsize, backend=backend,
        )  # fmt: skip
        self._arrays = arr
        super().__init__(base, cat.df)
        _BoxLikeMixin.__init__(self, categories, _splitby, color_by, hatch_by)
        base.with_edge(color=theme.get_theme().foreground_color)
        self._orient = orient
        self._numeric = cat._numeric

    @property
    def orient(self) -> Orientation:
        """Orientation of the violins."""
        return self._orient

    def move(self, shift: float = 0.0, autoscale: bool = True) -> Self:
        """Move the layer by the given shift."""
        base = self._base_layer
        data = base.data
        if self._orient.is_vertical:
            base.set_data(data.x + shift, data.y)
        else:
            base.set_data(data.x, data.y + shift)
        if autoscale and (canvas := self._canvas_ref()):
            canvas._autoscale_for_layer(self, pad_rel=0.025)
        return self

    def _get_arrays(self) -> list[np.ndarray]:
        return self._arrays

    def _set_estimation_values(self, est):
        if self.orient.is_vertical:
            self._base_layer.set_data(ydata=est)
        else:
            self._base_layer.set_data(xdata=est)

    def _set_error_values(self, err_low, err_high):
        mdata = self._base_layer.data
        if self.orient.is_vertical:
            self._base_layer.yerr.set_data(mdata.x, err_low, err_high)
        else:
            self._base_layer.xerr.set_data(err_low, err_high, mdata.y)

    def with_hover_text(self, text: str | list[str]) -> Self:
        """Set the hover tooltip text for the layer."""
        self.base.markers.with_hover_text(text)
        return self

    def with_hover_template(self, template: str) -> Self:
        """Set the hover tooltip template for the layer."""
        extra = {}
        for i, key in enumerate(self._splitby):
            extra[key] = [row[i] for row in self._categories]
        self.base.markers.with_hover_template(template, extra=extra)
        return self

    def _as_legend_item(self) -> _legend.LegendItemCollection:
        return _BoxLikeMixin._as_legend_item(self)


class DFBarPlot(
    _shared.DataFrameLayerWrapper[_lg.LabeledBars, _DF], _EstimatorMixin, Generic[_DF]
):
    def __init__(
        self,
        cat: CatIterator[_DF],
        value: str,
        color: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
        dodge: str | tuple[str, ...] | bool | None = None,
        name: str | None = None,
        orient: Orientation = Orientation.VERTICAL,
        capsize: float = 0.1,
        extent: float = 0.8,
        backend: str | Backend | None = None,
    ):
        _splitby, dodge = _shared.norm_dodge(
            cat.df, cat.offsets, color, hatch, dodge=dodge,
        )  # fmt: skip
        x, arr, categories = cat.prep_arrays(_splitby, value, dodge=dodge)
        _extent = cat.zoom_factor(dodge=dodge) * extent
        _capsize = cat.zoom_factor(dodge=dodge) * capsize
        color_by, hatch_by = _norm_color_hatch(color, hatch, cat.df)
        base = _lg.LabeledBars.from_arrays(
            x, arr, name=name, orient=orient, capsize=_capsize, extent=_extent,
            backend=backend,
        )  # fmt: skip
        self._arrays = arr
        super().__init__(base, cat.df)
        _BoxLikeMixin.__init__(self, categories, _splitby, color_by, hatch_by)
        base.with_edge(color=theme.get_theme().foreground_color)
        self._orient = orient
        self._numeric = cat._numeric

    @property
    def orient(self) -> Orientation:
        return self._base_layer.bars.orient

    def _get_arrays(self) -> list[np.ndarray]:
        return self._arrays

    def _set_estimation_values(self, est):
        if self.orient.is_vertical:
            self._base_layer.set_data(ydata=est)
        else:
            self._base_layer.set_data(xdata=est)

    def _set_error_values(self, err_low, err_high):
        mdata = self._base_layer.data
        if self.orient.is_vertical:
            self._base_layer.yerr.set_data(mdata.x, err_low, err_high)
        else:
            self._base_layer.xerr.set_data(err_low, err_high, mdata.y)

    def move(self, shift: float = 0.0, autoscale: bool = True) -> Self:
        """Move the layer by the given shift."""
        base = self._base_layer
        data = base.data
        if self._orient.is_vertical:
            base.set_data(data.x + shift, data.y)
        else:
            base.set_data(data.x, data.y + shift)
        if autoscale and (canvas := self._canvas_ref()):
            canvas._autoscale_for_layer(self, pad_rel=0.025)
        return self

    def with_hover_text(self, text: str | list[str]) -> Self:
        """Set the hover tooltip text for the layer."""
        self.base.bars.with_hover_text(text)
        return self

    def with_hover_template(self, template: str) -> Self:
        """Set the hover tooltip template for the layer."""
        extra = {}
        for i, key in enumerate(self._splitby):
            extra[key] = [row[i] for row in self._categories]
        self.base.bars.with_hover_template(template, extra=extra)
        return self

    def _as_legend_item(self) -> _legend.LegendItemCollection:
        return _BoxLikeMixin._as_legend_item(self)


_L0 = TypeVar("_L0", bound=Layer)
_L1 = TypeVar("_L1", bound=Layer)


def _combine_main_and_others(
    layer: _L0,
    incoming: _L1,
) -> _lg.MainAndOtherLayers[_L0, _L1]:
    if layer._group_layer_ref is None:
        return _lg.MainAndOtherLayers([layer, incoming], name=layer.name)
    group_layer = layer._group_layer_ref()
    if group_layer is None:
        raise ValueError("Parent layer group is deleted.")
    elif not isinstance(group_layer, _lg.MainAndOtherLayers):
        raise ValueError(f"Parent layer group is incorrect type {type(group_layer)}.")
    group_layer._insert(incoming)
    return group_layer



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Iterable,
    Sequence,
    TypeVar,
)

import numpy as np
from cmap import Color
from numpy.typing import NDArray
from psygnal import Signal, SignalGroup

from whitecanvas import protocols
from whitecanvas._exceptions import ReferenceDeletedError
from whitecanvas._signal import MouseMoveSignal, MouseSignal
from whitecanvas.types import (
    ColorType,
    LineStyle,
    Modifier,
    MouseButton,
    MouseEvent,
    MouseEventType,
    Point,
)
from whitecanvas.utils.normalize import arr_color

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.canvas._base import CanvasBase

_no_canvas = object()
_T = TypeVar("_T")


class AxisSignals(SignalGroup):
    """Signals emitted by an axis."""

    lim = Signal(tuple)


class _StrongRef(Generic[_T]):
    """Strong reference to an object."""

    def __init__(self, obj: _T):
        self._obj = obj

    def __call__(self) -> _T:
        return self._obj


class Namespace:
    _attrs: tuple[str, ...] = ()

    def __init__(self, canvas: CanvasBase | None = None):
        if canvas is not None:
            # This line *should* be an weak reference, but canvas is sometimes deleted
            # for some reason. Just use a strong reference for now.
            self._canvas_ref = _StrongRef(canvas)
        else:
            self._canvas_ref = _StrongRef(_no_canvas)
        self._instances: dict[int, Self] = {}

    def __get__(self, canvas, owner=None) -> Self:
        if canvas is None:
            return self
        while isinstance(canvas, Namespace):
            canvas = canvas._canvas_ref()
        id_ = id(canvas)
        if (ns := self._instances.get(id_)) is None:
            ns = self._instances[id_] = type(self)(canvas)
        return ns

    def _get_canvas(self) -> protocols.CanvasProtocol:
        l = self._canvas_ref()
        if l is None:
            raise ReferenceDeletedError("Canvas has been deleted.")
        elif l is _no_canvas:
            raise TypeError("No canvas is associated with the class itself.")
        return l._canvas()

    def _draw_canvas(self):
        if canvas := self._canvas_ref():
            canvas._draw_canvas()

    def __repr__(self) -> str:
        cname = type(self).__name__
        l = self._canvas_ref()
        if l is None:
            return f"<{cname} of deleted canvas>"
        elif l is _no_canvas:
            return f"<{cname}>"
        props = [f"canvas={l!r}"]
        for k in self._attrs:
            v = getattr(self, k)
            props.append(f"{k}={v!r}")
        return f"{cname}({', '.join(props)})"

    def update(self, d: dict[str, Any] = {}, **kwargs):
        values = dict(d, **kwargs)
        invalid_args = set(values) - set(self._attrs)
        if invalid_args:
            raise TypeError(f"Cannot set {invalid_args!r} on {type(self).__name__}")
        for k, v in values.items():
            setattr(self, k, v)


class _TextBoundNamespace(Namespace):
    _attrs = ("color", "size", "family", "visible")

    def _get_object(self) -> protocols.TextLabelProtocol:
        raise NotImplementedError

    @property
    def color(self):
        """Text color"""
        return self._get_object()._plt_get_color()

    @color.setter
    def color(self, color):
        self._get_object()._plt_set_color(np.array(Color(color)))

    @property
    def size(self) -> float:
        """Text font size"""
        return self._get_object()._plt_get_size()

    @size.setter
    def size(self, size: float):
        self._get_object()._plt_set_size(size)

    @property
    def family(self) -> str:
        """Text font family."""
        return self._get_object()._plt_get_fontfamily()

    @family.setter
    def family(self, font):
        self._get_object()._plt_set_fontfamily(font)

    @property
    def visible(self) -> bool:
        """Text visibility."""
        return self._get_object()._plt_get_visible()

    @visible.setter
    def visible(self, visible: bool):
        self._get_object()._plt_set_visible(visible)


class _TextLabelNamespace(_TextBoundNamespace):
    def __repr__(self) -> str:
        cname = type(self).__name__
        l = self._canvas_ref()
        if l is None:
            return f"<{cname} of deleted canvas>"
        elif l is _no_canvas:
            return f"<{cname}>"
        text = self.text
        color = self.color
        size = self.size
        family = self.family
        return f"{cname}({text=!r}, {color=!r}, {size=!r}, {family=!r})"

    @property
    def text(self) -> str:
        """Text content."""
        return self._get_object()._plt_get_text()

    @text.setter
    def text(self, text: str):
        self._get_object()._plt_set_text(text)

    def __set__(self, instance, value):
        # allow canvas.x.label = "X" as a shortcut for canvas.x.label.text = "X"
        obj = self.__get__(instance)
        if isinstance(value, str):
            obj.text = value
        elif isinstance(value, dict):
            obj.update(value)
        else:
            raise TypeError(f"Cannot set {value!r} to {type(obj)}.")


class _TicksNamespace(_TextBoundNamespace):
    def __repr__(self) -> str:
        cname = type(self).__name__
        l = self._canvas_ref()
        if l is None:
            return f"<{cname} of deleted canvas>"
        elif l is _no_canvas:
            return f"<{cname}>"
        pos, labels = self._get_object()._plt_get_tick_labels()
        pos = list(pos)
        color = self.color
        size = self.size
        family = self.family
        return f"{cname}({pos=!r}, {labels=}, {color=!r}, {size=!r}, {family=!r})"

    def _get_object(self) -> protocols.TicksProtocol:
        raise NotImplementedError

    @property
    def pos(self) -> NDArray[np.floating]:
        pos, _ = self._get_object()._plt_get_tick_labels()
        return np.asarray(pos)

    @property
    def labels(self) -> list[str]:
        _, labels = self._get_object()._plt_get_tick_labels()
        return labels

    def set_labels(self, pos: Iterable[float], labels: Iterable[str] | None = None):
        """
        Override tick labels.

        >>> canvas.x.ticks.set_labels([0, 1, 2], ["a", "b", "c"])

        Parameters
        ----------
        pos : iterable of float
            The positions of the ticks.
        labels : iterable of str, optional
            The label strings of the ticks. If None, the labels are set to the
            `pos` values.
        """
        _pos = list(pos)
        # test sorted
        if len(_pos) > 0 and np.any(np.diff(_pos) <= 0):
            raise ValueError(f"pos must be strictly increasing, got {pos}.")
        if labels is not None:
            _labels = [str(l) for l in labels]
        else:
            ndigits = int(np.log10(_pos[-1] - _pos[0])) + 1
            _labels = [str(round(p, ndigits)) for p in _pos]
        if len(_pos) != len(_labels):
            raise ValueError("pos and labels must have the same length.")
        self._get_object()._plt_override_labels(_pos, _labels)
        self._get_canvas()._plt_draw()

    def reset_labels(self) -> None:
        """Reset the tick labels to the default."""
        self._get_object()._plt_reset_override()
        self._get_canvas()._plt_draw()

    @property
    def rotation(self) -> float:
        """Tick label rotation in degrees."""
        return self._get_object()._plt_get_text_rotation()

    @rotation.setter
    def rotation(self, rotation: float):
        self._get_object()._plt_set_text_rotation(rotation)


class XTickNamespace(_TicksNamespace):
    def _get_object(self):
        return self._get_canvas()._plt_get_xticks()


class YTickNamespace(_TicksNamespace):
    def _get_object(self):
        return self._get_canvas()._plt_get_yticks()


class ZTickNamespace(_TicksNamespace):
    def _get_object(self):
        return self._get_canvas()._plt_get_zticks()


class TitleNamespace(_TextLabelNamespace):
    def _get_object(self):
        return self._get_canvas()._plt_get_title()


class XLabelNamespace(_TextLabelNamespace):
    def _get_object(self):
        return self._get_canvas()._plt_get_xlabel()


class YLabelNamespace(_TextLabelNamespace):
    def _get_object(self):
        return self._get_canvas()._plt_get_ylabel()


class ZLabelNamespace(_TextLabelNamespace):
    def _get_object(self):
        return self._get_canvas()._plt_get_zlabel()


class AxisNamespace(Namespace):
    events: AxisSignals
    _attrs = ("lim", "color", "flipped")

    def __init__(self, canvas: CanvasBase | None = None):
        super().__init__(canvas)
        self.events = AxisSignals()
        self._flipped = False

    def _get_object(self) -> protocols.AxisProtocol:
        raise NotImplementedError

    @property
    def lim(self) -> tuple[float, float]:
        """Limits of the axis."""
        return self._get_object()._plt_get_limits()

    @lim.setter
    def lim(self, lim: tuple[float, float]):
        low, high = lim
        if low >= high:
            a = type(self).__name__[0].lower()
            raise ValueError(
                f"low must be less than high, but got {lim!r}. If you "
                f"want to flip the axis, use `canvas.{a}.flipped = True`."
            )
        # Manually emit signal. This is needed when the plot backend is
        # implemented in JS (such as bokeh) and the python callback is not
        # enabled. Otherwise axis linking fails.
        with self.events.blocked():
            self._get_object()._plt_set_limits(lim)
        self.events.lim.emit(lim)
        return None

    @property
    def color(self):
        """Color of the axis."""
        return self._get_object()._plt_get_color()

    @color.setter
    def color(self, color):
        self._get_object()._plt_set_color(np.array(Color(color)))
        self._draw_canvas()

    @property
    def flipped(self) -> bool:
        """Return true if the axis is flipped."""
        return self._flipped

    @flipped.setter
    def flipped(self, flipped: bool):
        """Set the axis to be flipped."""
        if flipped != self._flipped:
            self._get_object()._plt_flip()
            self._flipped = flipped

    def set_gridlines(
        self,
        visible: bool = True,
        color: ColorType = "gray",
        width: float = 1.0,
        style: str | LineStyle = LineStyle.SOLID,
    ):
        color = arr_color(color)
        style = LineStyle(style)
        if width < 0:
            raise ValueError("width must be non-negative.")
        self._get_object()._plt_set_grid_state(visible, color, width, style)


class XAxisNamespace(AxisNamespace):
    label = XLabelNamespace()
    ticks = XTickNamespace()

    def _get_object(self):
        return self._get_canvas()._plt_get_xaxis()


class YAxisNamespace(AxisNamespace):
    label = YLabelNamespace()
    ticks = YTickNamespace()

    def _get_object(self):
        return self._get_canvas()._plt_get_yaxis()


class ZAxisNamespace(AxisNamespace):
    label = ZLabelNamespace()
    ticks = ZTickNamespace()

    def _get_object(self):
        return self._get_canvas()._plt_get_zaxis()


class MouseNamespace(AxisNamespace):
    """Namespace that contains the mouse events."""

    clicked = MouseSignal(object)
    """Signal emitted when a mouse button is clicked."""

    moved = MouseMoveSignal()
    """Signal emitted when the mouse is moved."""

    double_clicked = MouseSignal(object)
    """Signal emitted when a mouse button is double-clicked."""

    @property
    def enabled(self) -> bool:
        """Return whether pan/zoom is enabled."""
        return self._get_canvas()._plt_get_mouse_enabled()

    @enabled.setter
    def enabled(self, enabled: bool):
        self._get_canvas()._plt_set_mouse_enabled(enabled)

    def emulate_click(
        self,
        position: tuple[float, float],
        *,
        button: str | MouseButton = MouseButton.LEFT,
        modifiers: str | Modifier | Sequence[str | Modifier] = (),
    ) -> None:
        """Emulate a mouse press event."""
        ev = MouseEvent(
            MouseButton(button),
            _norm_modifiers(modifiers),
            Point(*position),
            MouseEventType.PRESS,
        )
        self.clicked.emit(ev)
        return None

    def emulate_double_click(
        self,
        position: tuple[float, float],
        *,
        button: str | MouseButton = MouseButton.LEFT,
        modifiers: str | Modifier | Sequence[str | Modifier] = (),
    ) -> None:
        """Emulate a mouse double-click event."""
        ev = MouseEvent(
            MouseButton(button),
            _norm_modifiers(modifiers),
            Point(*position),
            MouseEventType.DOUBLE_CLICK,
        )
        self.double_clicked.emit(ev)

    def emulate_hover(
        self,
        positions: Sequence[tuple[float, float]],
        *,
        modifiers: str | Modifier | Sequence[str | Modifier] = (),
    ) -> None:
        """Emulate a mouse move event."""
        _modifiers = _norm_modifiers(modifiers)

        for pos in positions:
            ev = MouseEvent(
                MouseButton.NONE,
                _modifiers,
                Point(*pos),
                MouseEventType.MOVE,
            )
            self.moved.emit(ev)
        return None

    def emulate_drag(
        self,
        positions: Sequence[tuple[float, float]],
        *,
        button: str | MouseButton = MouseButton.LEFT,
        modifiers: str | Modifier | Sequence[str | Modifier] = (),
    ):
        """Emulate a mouse press-move-release event."""
        _modifiers = _norm_modifiers(modifiers)

        ev = MouseEvent(
            MouseButton(button),
            _modifiers,
            Point(*positions[0]),
            MouseEventType.PRESS,
        )
        self.moved.emit(ev)

        for pos in positions[1:]:
            ev = MouseEvent(
                MouseButton(button),
                _modifiers,
                Point(*pos),
                MouseEventType.MOVE,
            )
            self.moved.emit(ev)

        ev = MouseEvent(
            MouseButton(button),
            _modifiers,
            Point(*positions[-1]),
            MouseEventType.RELEASE,
        )
        self.moved.emit(ev)
        return None


def _norm_modifiers(modifiers) -> tuple[Modifier, ...]:
    if isinstance(modifiers, str):
        _modifiers = (Modifier(modifiers),)
    elif isinstance(modifiers, Modifier):
        _modifiers = [modifiers]
    else:
        _modifiers = tuple(Modifier(m) for m in modifiers)
    return _modifiers



from __future__ import annotations

import weakref
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Generic,
    Literal,
    NamedTuple,
    Sequence,
    TypeVar,
    overload,
)

import numpy as np
from numpy.typing import NDArray
from psygnal import Signal

from whitecanvas.canvas import CanvasBase
from whitecanvas.layers import Layer, Line, Rects, Spans
from whitecanvas.tools._polygon_utils import is_in_polygon
from whitecanvas.types import (
    ColorType,
    LineStyle,
    Modifier,
    MouseButton,
    MouseEvent,
    MouseEventType,
    Point,
    Rect,
    XYData,
    _Void,
)

if TYPE_CHECKING:
    from whitecanvas.layers._mixin import ConstEdge, ConstFace

    _MouseButton = Literal["left", "middle", "right", "back", "forward"] | MouseButton
    _Modifier = Literal["shift", "ctrl", "alt", "meta"] | Modifier

_L = TypeVar("_L", bound=Layer)
_void = _Void()


class SelectionToolBase(ABC, Generic[_L]):
    changed = Signal(object)
    """Emitted when the selection is changed."""

    cleared = Signal()
    """Emitted when the selection is cleared by user."""

    def __init__(
        self,
        canvas: CanvasBase,
        buttons: list[MouseButton],
        modifiers: list[Modifier],
        tracking: bool = False,
    ):
        self._canvas_ref = weakref.ref(canvas)
        self._valid_buttons = set(buttons)
        self._modifiers = set(modifiers)
        self._layer = self._create_layer()
        self._tracking = tracking
        self._persist = True
        self._enabled = True
        canvas.mouse.moved.connect(self.callback)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self._canvas_ref() is not None:
            self.disconnect()

    def _canvas(self) -> CanvasBase:
        canvas = self._canvas_ref()
        if canvas is None:
            raise RuntimeError("Canvas has been deleted")
        return canvas

    @abstractmethod
    def _create_layer(self) -> _L: ...

    @abstractmethod
    def _update_layer(
        self,
        start: tuple[float, float],
        now: tuple[float, float],
    ): ...

    def _on_press(self, start: tuple[float, float]):
        pass

    @property
    @abstractmethod
    def selection(self):
        """This method returns the values that represents current selections."""

    @property
    def persist(self) -> bool:
        """Whether the selection persists after mouse release."""
        return self._persist

    @persist.setter
    def persist(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("persist must be a boolean value.")
        self._persist = value

    def callback(self, e: MouseEvent):
        """The callback function that is called when mouse is moved."""
        if not self._enabled:
            return
        if e.button not in self._valid_buttons or set(e.modifiers) != self._modifiers:
            return
        canvas = self._canvas()
        pos_start = e.pos
        self._on_press(pos_start)
        yield
        if self._layer in self._canvas().layers:
            self.clear_selection()
        with canvas.autoscale_context(enabled=False):
            canvas.add_layer(self._layer)
        dragged = False
        while e.type is not MouseEventType.RELEASE:
            self._update_layer(pos_start, e.pos)
            yield
            if self._tracking:
                self.changed.emit(self.selection)
            dragged = True

        if dragged:
            if not self._tracking:
                self.changed.emit(self.selection)
            if not self._persist:
                with self.cleared.blocked():
                    self.clear_selection()
        else:
            # clicked
            if self._persist:
                self.clear_selection()

    def clear_selection(self):
        """Clear the current selection."""
        self._canvas().layers.remove(self._layer)
        self.cleared.emit()

    def disconnect(self):
        """Disconnect the tool from the canvas."""
        self._canvas().mouse.moved.disconnect(self.callback)
        if self._layer in self._canvas().layers:
            self.clear_selection()
        return None

    @property
    def enabled(self) -> bool:
        """Whether the tool is enabled."""
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool):
        self._enabled = bool(value)
        if not self._enabled:
            self.clear_selection()


class RectSelectionTool(SelectionToolBase[Rects]):
    def _create_layer(self) -> Rects:
        layer = Rects([[0, 1, 0, 1]], color="blue", alpha=0.25).with_edge(
            width=2, alpha=0.4
        )
        layer.visible = False
        return layer

    def _update_layer(
        self,
        start: tuple[float, float],
        now: tuple[float, float],
    ):
        x0, y0 = start
        x1, y1 = now
        self._layer.data = np.array([[x0, x1, y0, y1]])
        self._layer.visible = True

    @property
    def selection(self) -> Rect:
        return self._layer.rects[0]

    @property
    def face(self) -> ConstFace:
        """Face color of the selection span."""
        return self._layer.face

    @property
    def edge(self) -> ConstEdge:
        """Edge color of the selection span."""
        return self._layer.edge

    @overload
    def contains_point(self, point: tuple[float, float], /) -> bool: ...
    @overload
    def contains_point(self, x: float, y: float, /) -> bool: ...

    def contains_point(self, *args) -> bool:
        x, y = _point_to_xy(*args)
        sel = self.selection
        return sel.left <= x <= sel.right and sel.bottom <= y <= sel.top

    def contains_points(self, points: XYData | NDArray[np.number]) -> NDArray[np.bool_]:
        points = _atleast_2d(points)
        sel = self.selection
        if points.ndim == 2 and points.shape[1] == 2:
            xs = points[:, 0]
            ys = points[:, 1]
            return (
                (sel.left <= xs)
                & (xs <= sel.right)
                & (sel.bottom <= ys)
                & (ys <= sel.top)
            )
        else:
            raise ValueError("points must be (2,) or (N, 2) array.")


class LineSelection(NamedTuple):
    start: Point
    end: Point


class LineSelectionTool(SelectionToolBase[Line]):
    def _create_layer(self) -> Line:
        return Line(np.array([]), np.array([]), width=2, color="blue", alpha=0.4)

    def _update_layer(
        self,
        start: tuple[float, float],
        now: tuple[float, float],
    ):
        x0, y0 = start
        x1, y1 = now
        self._layer.data = np.array([x0, x1]), np.array([y0, y1])

    @property
    def selection(self) -> LineSelection:
        xs, ys = self._layer.data
        return LineSelection(Point(xs[0], ys[0]), Point(xs[1], ys[1]))

    def update(
        self,
        *,
        color: ColorType | _Void = _void,
        width: float | _Void = _void,
        style: str | LineStyle | _Void = _void,
        alpha: float | _Void = _void,
    ):
        self._layer.update(color=color, width=width, style=style, alpha=alpha)

    @property
    def color(self) -> NDArray[np.float32]:
        """Color of the selection line."""
        return self._layer.color

    @color.setter
    def color(self, color: ColorType):
        self._layer.color = color

    @property
    def width(self) -> float:
        """Width of the selection line."""
        return self._layer.width

    @width.setter
    def width(self, width: float):
        self._layer.width = width

    @property
    def style(self) -> LineStyle:
        """Style of the selection line."""
        return self._layer.style

    @style.setter
    def style(self, style: str | LineStyle):
        self._layer.style = style

    @property
    def alpha(self) -> float:
        """Alpha channel of the selection line."""
        return self._layer.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._layer.alpha = alpha


class SpanSelection(NamedTuple):
    start: float
    end: float


class _SpanSelectionTool(SelectionToolBase[Spans]):
    @property
    def selection(self) -> SpanSelection:
        span = self._layer.data[0]
        return SpanSelection(*sorted(span))

    @property
    def face(self) -> ConstFace:
        """Face color of the selection span."""
        return self._layer.face

    @property
    def edge(self) -> ConstEdge:
        """Edge color of the selection span."""
        return self._layer.edge


class XSpanSelectionTool(_SpanSelectionTool):
    def _create_layer(self) -> Spans:
        layer = Spans([[0, 1]], orient="vertical", color="red", alpha=0.25)
        layer.visible = False
        return layer

    def _update_layer(
        self,
        start: tuple[float, float],
        now: tuple[float, float],
    ):
        x0, _ = start
        x1, _ = now
        self._layer.data = np.array([[x0, x1]], dtype=np.float32)
        self._layer.visible = True

    @overload
    def contains_point(self, point: tuple[float, float], /) -> bool: ...
    @overload
    def contains_point(self, x: float, y: float, /) -> bool: ...

    def contains_point(self, *args) -> bool:
        x, _ = _point_to_xy(*args)
        sel = self.selection
        return sel.start <= x <= sel.end

    def contains_points(self, points: XYData | NDArray[np.number]) -> NDArray[np.bool_]:
        points = _atleast_2d(points)
        sel = self.selection
        if points.ndim == 2 and points.shape[1] == 2:
            xs = points[:, 0]
            return (sel.start <= xs) & (xs <= sel.end)
        else:
            raise ValueError("points must be (2,) or (N, 2) array.")


class YSpanSelectionTool(_SpanSelectionTool):
    def _create_layer(self) -> Spans:
        layer = Spans([[0, 1]], orient="horizontal", color="red", alpha=0.25)
        layer.visible = False
        return layer

    def _update_layer(
        self,
        start: tuple[float, float],
        now: tuple[float, float],
    ):
        _, y0 = start
        _, y1 = now
        self._layer.data = np.array([[y0, y1]], dtype=np.float32)
        self._layer.visible = True

    @overload
    def contains_point(self, point: tuple[float, float], /) -> bool: ...
    @overload
    def contains_point(self, x: float, y: float, /) -> bool: ...

    def contains_point(self, *args) -> bool:
        _, y = _point_to_xy(*args)
        sel = self.selection
        return sel.start <= y <= sel.end

    def contains_points(self, points: XYData | NDArray[np.number]) -> NDArray[np.bool_]:
        points = _atleast_2d(points)
        sel = self.selection
        if points.ndim == 2 and points.shape[1] == 2:
            ys = points[:, 1]
            return (sel.start <= ys) & (ys <= sel.end)
        else:
            raise ValueError("points must be (2,) or (N, 2) array.")


class LassoSelectionTool(LineSelectionTool):
    def _update_layer(
        self,
        start: tuple[float, float],
        now: tuple[float, float],
    ):
        x1, y1 = now
        current = self._layer.data
        xs = np.concatenate([current.x, [x1]])
        ys = np.concatenate([current.y, [y1]])
        self._layer.data = xs, ys

    @property
    def selection(self) -> XYData:
        return self._layer.data

    def _on_press(self, start: tuple[float, float]):
        self._layer.data = np.array([start[0]]), np.array([start[1]])

    def close_path(self, emit: bool = False):
        """
        Close the path by connecting the last point to the first point.

        Parameters
        ----------
        emit : bool, default False
            If True, the changed signal is emitted after closing the path.
        """
        current = self._layer.data
        if len(current.x) > 2:
            xs = np.concatenate([current.x, [current.x[0]]])
            ys = np.concatenate([current.y, [current.y[0]]])
            self._layer.data = xs, ys
        if emit:
            self.changed.emit(self.selection)
        return

    @overload
    def contains_point(self, point: tuple[float, float], /) -> bool: ...
    @overload
    def contains_point(self, x: float, y: float, /) -> bool: ...

    def contains_point(self, *args) -> bool:
        x, y = _point_to_xy(*args)
        poly = self._layer.data
        return is_in_polygon(np.array([[x, y]]), poly.stack())[0]

    def contains_points(self, points: XYData | NDArray[np.number]) -> NDArray[np.bool_]:
        points = _atleast_2d(points)
        poly = self._layer.data
        if points.ndim == 2 and points.shape[1] == 2:
            return is_in_polygon(points, poly.stack())
        else:
            raise ValueError("points must be (2,) or (N, 2) array.")


class PolygonSelectionTool(LassoSelectionTool):
    def callback(self, e: MouseEvent):
        """The callback function that is called when mouse is moved."""
        if not self._enabled:
            return
        if e.button not in self._valid_buttons or set(e.modifiers) != self._modifiers:
            return
        canvas = self._canvas()
        pos_start = e.pos
        self._on_press(pos_start)
        yield
        if self._layer in self._canvas().layers:
            self.clear_selection()
            return
        with canvas.autoscale_context(enabled=False):
            canvas.add_layer(self._layer)

        while True:
            while e.type is not MouseEventType.RELEASE:
                yield  # dragging
            self._update_layer(pos_start, e.pos)
            yield
            while e.button is MouseButton.NONE:
                cur_data = self._layer.data
                x0, y0 = e.pos
                xs = np.concatenate([cur_data.x[:-1], [x0]])
                ys = np.concatenate([cur_data.y[:-1], [y0]])
                self._layer.data = xs, ys
                yield
            if e.type is MouseEventType.DOUBLE_CLICK:
                break
            elif e.button in self._valid_buttons:
                if e.type is MouseEventType.PRESS:
                    self._update_layer(pos_start, e.pos)
                    if self._tracking:
                        self.changed.emit(self.selection)
            elif e.type is MouseEventType.PRESS:
                break
            yield
        yield
        if not self._tracking:
            self.changed.emit(self.selection)
        if not self._persist:
            with self.cleared.blocked():
                self.clear_selection()


def _norm_input(
    buttons: _MouseButton | Sequence[_MouseButton] = "left",
    modifiers: _Modifier | Sequence[_Modifier] | None = None,
):
    if isinstance(buttons, (str, MouseButton)):
        buttons = [buttons]
    _buttons = [MouseButton(btn) for btn in buttons]
    if MouseButton.NONE in _buttons:
        raise ValueError("MouseButton.NONE is not allowed.")
    if modifiers is None:
        modifiers = []
    elif isinstance(modifiers, (str, Modifier)):
        modifiers = [modifiers]
    _modifiers = [Modifier(mod) for mod in modifiers]
    return _buttons, _modifiers


def _atleast_2d(points: NDArray[np.number]) -> NDArray[np.number]:
    if isinstance(points, XYData):
        return points.stack()
    return np.atleast_2d(points)


def _point_to_xy(*args) -> tuple[float, float]:
    if len(args) == 1:
        x, y = args[0]
    else:
        x, y = args
    return x, y


def line_selector(
    canvas: CanvasBase,
    buttons: _MouseButton | Sequence[_MouseButton] = "left",
    modifiers: _Modifier | Sequence[_Modifier] | None = None,
    *,
    tracking: bool = False,
) -> LineSelectionTool:
    """
    Create a line selector tool with given settings.

    A line selector emits a LineSelection object when a line is drawn.
    A selection tool is constructed by specifying the canvas to attach the tool.

    >>> canvas = new_canvas("matplotlib:qt")
    >>> tool = line_selector(canvas)

    Use `buttons` and `modifiers` to specify how to trigger the tool.

    >>> tool = line_selector(canvas, buttons="right", modifiers="ctrl")

    Parameters
    ----------
    canvas : CanvasBase
        The canvas to which the tool is attached.
    buttons : MouseButton or Sequence[MouseButton], default "left"
        The mouse buttons that can trigger the tool.
    modifiers : Modifier or Sequence[Modifier], optional
        The modifier keys that must be pressed to trigger the tool.
    tracking : bool, default False
        If True, the tool emits the changed signal while dragging. Otherwise, it emits
        the signal only when dragging is finished.
    """
    _buttons, _modifiers = _norm_input(buttons, modifiers)
    return LineSelectionTool(canvas, _buttons, _modifiers, tracking=tracking)


def rect_selector(
    canvas: CanvasBase,
    buttons: _MouseButton | Sequence[_MouseButton] = "left",
    modifiers: _Modifier | Sequence[_Modifier] | None = None,
    *,
    tracking: bool = False,
) -> RectSelectionTool:
    """
    Create a rectangle selector tool with given settings.

    A rectangle selector emits a Rect object when a rectangle is drawn.
    A selection tool is constructed by specifying the canvas to attach the tool.

    >>> canvas = new_canvas("matplotlib:qt")
    >>> tool = rect_selector(canvas)

    Use `buttons` and `modifiers` to specify how to trigger the tool.

    >>> tool = rect_selector(canvas, buttons="right", modifiers="ctrl")

    Parameters
    ----------
    canvas : CanvasBase
        The canvas to which the tool is attached.
    buttons : MouseButton or Sequence[MouseButton], default "left"
        The mouse buttons that can trigger the tool.
    modifiers : Modifier or Sequence[Modifier], optional
        The modifier keys that must be pressed to trigger the tool.
    tracking : bool, default False
        If True, the tool emits the changed signal while dragging. Otherwise, it emits
        the signal only when dragging is finished.
    """
    _buttons, _modifiers = _norm_input(buttons, modifiers)
    return RectSelectionTool(canvas, _buttons, _modifiers, tracking=tracking)


def xspan_selector(
    canvas: CanvasBase,
    buttons: _MouseButton | Sequence[_MouseButton] = "left",
    modifiers: _Modifier | Sequence[_Modifier] | None = None,
    *,
    tracking: bool = False,
) -> XSpanSelectionTool:
    """
    Create a x-span selector tool with given settings.

    A x-span selector emits a SpanSelection object (tuple of start and end) when a span
    is drawn.
    A selection tool is constructed by specifying the canvas to attach the tool.

    >>> canvas = new_canvas("matplotlib:qt")
    >>> tool = xspan_selector(canvas)

    Use `buttons` and `modifiers` to specify how to trigger the tool.

    >>> tool = line_selector(canvas, buttons="right", modifiers="ctrl")

    Parameters
    ----------
    canvas : CanvasBase
        The canvas to which the tool is attached.
    buttons : MouseButton or Sequence[MouseButton], default "left"
        The mouse buttons that can trigger the tool.
    modifiers : Modifier or Sequence[Modifier], optional
        The modifier keys that must be pressed to trigger the tool.
    tracking : bool, default False
        If True, the tool emits the changed signal while dragging. Otherwise, it emits
        the signal only when dragging is finished.
    """
    _buttons, _modifiers = _norm_input(buttons, modifiers)
    return XSpanSelectionTool(canvas, _buttons, _modifiers, tracking=tracking)


def yspan_selector(
    canvas: CanvasBase,
    buttons: _MouseButton | Sequence[_MouseButton] = "left",
    modifiers: _Modifier | Sequence[_Modifier] | None = None,
    *,
    tracking: bool = False,
) -> YSpanSelectionTool:
    """
    Create a line selector tool with given settings.

    A y-span selector emits a SpanSelection object (tuple of start and end) when a span
    is drawn.
    A selection tool is constructed by specifying the canvas to attach the tool.

    >>> canvas = new_canvas("matplotlib:qt")
    >>> tool = yspan_selector(canvas)

    Use `buttons` and `modifiers` to specify how to trigger the tool.

    >>> tool = line_selector(canvas, buttons="right", modifiers="ctrl")

    Parameters
    ----------
    canvas : CanvasBase
        The canvas to which the tool is attached.
    buttons : MouseButton or Sequence[MouseButton], default "left"
        The mouse buttons that can trigger the tool.
    modifiers : Modifier or Sequence[Modifier], optional
        The modifier keys that must be pressed to trigger the tool.
    tracking : bool, default False
        If True, the tool emits the changed signal while dragging. Otherwise, it emits
        the signal only when dragging is finished.
    """
    _buttons, _modifiers = _norm_input(buttons, modifiers)
    return YSpanSelectionTool(canvas, _buttons, _modifiers, tracking=tracking)


def lasso_selector(
    canvas: CanvasBase,
    buttons: _MouseButton | Sequence[_MouseButton] = "left",
    modifiers: _Modifier | Sequence[_Modifier] | None = None,
    *,
    tracking: bool = False,
) -> LassoSelectionTool:
    """
    Create a Lasso selector tool with given settings.

    A Lasso selector emits a XYData object by freehand drawing.
    A selection tool is constructed by specifying the canvas to attach the tool.

    >>> canvas = new_canvas("matplotlib:qt")
    >>> tool = lasso_selector(canvas)

    Use `buttons` and `modifiers` to specify how to trigger the tool.

    >>> tool = lasso_selector(canvas, buttons="right", modifiers="ctrl")

    Parameters
    ----------
    canvas : CanvasBase
        The canvas to which the tool is attached.
    buttons : MouseButton or Sequence[MouseButton], default "left"
        The mouse buttons that can trigger the tool.
    modifiers : Modifier or Sequence[Modifier], optional
        The modifier keys that must be pressed to trigger the tool.
    tracking : bool, default False
        If True, the tool emits the changed signal while dragging. Otherwise, it emits
        the signal only when dragging is finished.
    """
    _buttons, _modifiers = _norm_input(buttons, modifiers)
    return LassoSelectionTool(canvas, _buttons, _modifiers, tracking=tracking)


def polygon_selector(
    canvas: CanvasBase,
    buttons: _MouseButton | Sequence[_MouseButton] = "left",
    modifiers: _Modifier | Sequence[_Modifier] | None = None,
    *,
    tracking: bool = False,
) -> LassoSelectionTool:
    """
    Create a polygon selector tool with given settings.

    A polygon selector emits a XYData object by freehand drawing.
    A selection tool is constructed by specifying the canvas to attach the tool.

    >>> canvas = new_canvas("matplotlib:qt")
    >>> tool = polygon_selector(canvas)

    Use `buttons` and `modifiers` to specify how to trigger the tool.

    >>> tool = polygon_selector(canvas, buttons="right", modifiers="ctrl")

    Parameters
    ----------
    canvas : CanvasBase
        The canvas to which the tool is attached.
    buttons : MouseButton or Sequence[MouseButton], default "left"
        The mouse buttons that can trigger the tool.
    modifiers : Modifier or Sequence[Modifier], optional
        The modifier keys that must be pressed to trigger the tool.
    tracking : bool, default False
        If True, the tool emits the changed signal while dragging. Otherwise, it emits
        the signal only when dragging is finished.
    """
    _buttons, _modifiers = _norm_input(buttons, modifiers)
    return PolygonSelectionTool(canvas, _buttons, _modifiers, tracking=tracking)



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Generic, TypeVar

import numpy as np
from numpy.typing import NDArray

from whitecanvas.layers._base import Layer, PrimitiveLayer
from whitecanvas.protocols import BaseProtocol

_P = TypeVar("_P", bound=BaseProtocol)
_T = TypeVar("_T")


class Layer3D(Layer):
    def __init__(self, name: str | None = None):
        super().__init__(name=name)
        self._z_hint = None

    def bbox_hint(self) -> NDArray[np.float64]:
        return np.array([0, 0, 0, 1, 1, 1], dtype=np.float64)


class PrimitiveLayer3D(PrimitiveLayer[_P], Layer3D):
    def bbox_hint(self) -> NDArray[np.float64]:
        """Return the bounding box hint (xmin, xmax, ymin, ymax, zmin, zmax)."""
        if self._x_hint is None:
            _x = (np.nan, np.nan)
        else:
            _x = self._x_hint
        if self._y_hint is None:
            _y = (np.nan, np.nan)
        else:
            _y = self._y_hint
        if self._z_hint is None:
            _z = (np.nan, np.nan)
        else:
            _z = self._z_hint
        return np.array(_x + _y + _z, dtype=np.float64)


class DataBoundLayer3D(PrimitiveLayer3D[_P], Generic[_P, _T]):
    @abstractmethod
    def _get_layer_data(self) -> _T:
        """Get the data for this layer."""

    @abstractmethod
    def _set_layer_data(self, data: _T):
        """Set the data for this layer."""

    def _norm_layer_data(self, data: Any) -> _T:
        """Normalize the data for this layer."""
        return data

    @property
    def data(self) -> _T:
        """Data for this layer."""
        return self._get_layer_data()

    @data.setter
    def data(self, data):
        """Set the data for this layer."""
        data_normed = self._norm_layer_data(data)
        self._set_layer_data(data_normed)
        self.events.data.emit(data_normed)



from __future__ import annotations

from dataclasses import dataclass
from typing import NamedTuple

from whitecanvas.types._enums import Modifier, MouseButton, MouseEventType


class Point(NamedTuple):
    x: float
    y: float


@dataclass
class MouseEvent:
    button: MouseButton
    modifiers: tuple[Modifier, ...]
    pos: Point
    type: MouseEventType

    def __post_init__(self):
        self.pos = Point(*self.pos)  # normalize

    def update(self, other: MouseEvent):
        self.button = other.button
        self.modifiers = other.modifiers
        self.pos = other.pos
        self.type = other.type

    def _repr_simple(self):
        modifiers = "+".join(mod.name for mod in self.modifiers)
        return (
            f"button={self.button.name!r}, "
            f"{modifiers=!r}, pos={self.pos!r}, type={self.type.name!r}"
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from psygnal import Signal

from whitecanvas.backend import Backend
from whitecanvas.layers import _legend
from whitecanvas.layers._base import DataBoundLayer
from whitecanvas.layers._mixin import FaceEdgeMixin, FaceEdgeMixinEvents
from whitecanvas.layers._sizehint import xyy_size_hint
from whitecanvas.protocols import BandProtocol
from whitecanvas.types import (
    ArrayLike1D,
    ColorType,
    Hatch,
    Orientation,
    OrientationLike,
    XYYData,
)
from whitecanvas.utils.normalize import as_array_1d

if TYPE_CHECKING:
    from typing_extensions import Self


class BandEvents(FaceEdgeMixinEvents):
    clicked = Signal()


class Band(DataBoundLayer[BandProtocol, XYYData], FaceEdgeMixin):
    events: BandEvents
    _events_class = BandEvents

    def __init__(
        self,
        t: ArrayLike1D,
        edge_low: ArrayLike1D,
        edge_high: ArrayLike1D,
        orient: OrientationLike = "vertical",
        *,
        name: str | None = None,
        color: ColorType = "blue",
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
        backend: Backend | str | None = None,
    ):
        ori = Orientation.parse(orient)
        x = as_array_1d(t)
        y0 = as_array_1d(edge_low)
        y1 = as_array_1d(edge_high)
        if x.size != y0.size or x.size != y1.size:
            raise ValueError(
                "Expected xdata, ydata0, ydata1 to have the same size, "
                f"got {x.size}, {y0.size}, {y1.size}"
            )
        super().__init__(name=name if name is not None else "Band")
        FaceEdgeMixin.__init__(self)
        self._backend = self._create_backend(Backend(backend), x, y0, y1, ori)
        self._orient = ori
        self.face.update(color=color, alpha=alpha, hatch=hatch)
        self._x_hint, self._y_hint = xyy_size_hint(x, y0, y1, ori)
        self._band_type = "band"
        self.edge.width = 0.0
        self._init_events()
        self._backend._plt_connect_pick_event(self.events.clicked.emit)

    @property
    def orient(self) -> Orientation:
        """Orientation of the band."""
        return self._orient

    def _get_layer_data(self) -> XYYData:
        """Current data of the layer."""
        if self._orient.is_vertical:
            x, y0, y1 = self._backend._plt_get_vertical_data()
        else:
            x, y0, y1 = self._backend._plt_get_horizontal_data()
        return XYYData(x, y0, y1)

    def _norm_layer_data(self, data: Any) -> XYYData:
        t0, y0, y1 = self.data
        t, edge_low, edge_high = data
        if t is not None:
            t0 = t
        if edge_low is not None:
            y0 = as_array_1d(edge_low)
        if edge_high is not None:
            y1 = as_array_1d(edge_high)
        if t0.size != y0.size or t0.size != y1.size:
            raise ValueError(
                "Expected data to have the same size,"
                f"got {t0.size}, {y0.size}, {y1.size}"
            )
        return XYYData(t0, y0, y1)

    def _set_layer_data(self, data: XYYData):
        t0, y0, y1 = data
        if self._orient.is_vertical:
            self._backend._plt_set_vertical_data(t0, y0, y1)
        else:
            self._backend._plt_set_horizontal_data(t0, y0, y1)
        self._x_hint, self._y_hint = xyy_size_hint(t0, y0, y1, self.orient)

    def set_data(
        self,
        t: ArrayLike1D | None = None,
        edge_low: ArrayLike1D | None = None,
        edge_high: ArrayLike1D | None = None,
    ):
        self.data = t, edge_low, edge_high

    def with_hover_text(self, text: str) -> Self:
        """Add hover text to the data points."""
        self._backend._plt_set_hover_text(str(text))
        return self

    def _as_legend_item(self) -> _legend.BarLegendItem:
        face = _legend.FaceInfo(self.face.color, self.face.hatch)
        edge = _legend.EdgeInfo(self.edge.color, self.edge.width, self.edge.style)
        return _legend.BarLegendItem(face, edge)



from __future__ import annotations

import tempfile
import warnings
from pathlib import Path
from typing import Generic, TypeVar

import bokeh.events as bk_events
import bokeh.models as bk_models
import numpy as np
from numpy.typing import NDArray

from whitecanvas.protocols import BaseProtocol
from whitecanvas.types import Hatch, LineStyle, Symbol
from whitecanvas.utils.normalize import arr_color, hex_color

_M = TypeVar("_M", bound=bk_models.Model)


class BokehLayer(BaseProtocol, Generic[_M]):
    _model: _M
    _data: bk_models.ColumnDataSource


class HeteroLayer(BokehLayer[_M]):
    ##### LayerProtocol #####
    def _plt_get_visible(self) -> bool:
        return self._visible

    def _plt_set_visible(self, visible: bool):
        if visible:
            self._model.line_color = "edge_color"
            self._model.fill_color = "face_color"
        else:
            self._model.line_color = "#00000000"
            self._model.fill_color = "#00000000"
        self._visible = visible

    ##### HasFace protocol #####

    def _plt_get_face_color(self) -> NDArray[np.float32]:
        colors = [arr_color(c) for c in self._data.data["face_color"]]
        if len(colors) == 0:
            return np.zeros((0, 4), dtype=np.float32)
        return np.stack(colors, axis=0)

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        if color.ndim == 1:
            color = [hex_color(color)] * self._plt_get_ndata()
        else:
            color = [hex_color(c) for c in color]
        self._data.data["face_color"] = color

    def _plt_get_face_hatch(self) -> list[Hatch]:
        return [from_bokeh_hatch(p) for p in self._data.data["pattern"]]

    def _plt_set_face_hatch(self, pattern: Hatch | list[Hatch]):
        if isinstance(pattern, Hatch):
            ptn = [to_bokeh_hatch(pattern)] * self._plt_get_ndata()
        else:
            ptn = [to_bokeh_hatch(p) for p in pattern]
        self._data.data["pattern"] = ptn

    ##### HasEdges protocol #####

    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        return self._data.data["width"]

    def _plt_set_edge_width(self, width: float):
        if np.isscalar(width):
            width = np.full(self._plt_get_ndata(), width)
        self._data.data["width"] = width

    def _plt_get_edge_style(self) -> list[LineStyle]:
        return [from_bokeh_line_style(d) for d in self._data.data["style"]]

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        if isinstance(style, LineStyle):
            style = [style] * self._plt_get_ndata()
        val = [to_bokeh_line_style(s) for s in style]
        self._data.data["style"] = val

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        colors = [arr_color(c) for c in self._data.data["edge_color"]]
        if len(colors) == 0:
            return np.zeros((0, 4), dtype=np.float32)
        return np.stack(colors, axis=0)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        if color.ndim == 1:
            color = [hex_color(color)] * self._plt_get_ndata()
        else:
            color = [hex_color(c) for c in color]
        self._data.data["edge_color"] = color


class SupportsMouseEvents:
    def _plt_connect_pick_event(self, callback):
        # TODO: not working
        for event in [bk_events.Tap, bk_events.Press, bk_events.PanEnd]:
            self._model.on_event(event, lambda e: callback(e.indices))

    def _plt_set_hover_text(self, text: list[str]):
        self._data.data["hovertexts"] = text


def to_bokeh_line_style(style: LineStyle) -> str:
    if style is LineStyle.SOLID:
        return "solid"
    elif style is LineStyle.DASH:
        return "dashed"
    elif style is LineStyle.DOT:
        return "dotted"
    elif style is LineStyle.DASH_DOT:
        return "dashdot"


def from_bokeh_line_style(style: str) -> LineStyle:
    if style == "solid":
        return LineStyle.SOLID
    elif style == "dashed":
        return LineStyle.DASH
    elif style == "dotted":
        return LineStyle.DOT
    elif style in ("dashdot", "dotdash"):
        return LineStyle.DASH_DOT
    else:
        return LineStyle.SOLID


def to_bokeh_symbol(symbol: Symbol) -> tuple[str, float]:
    return _SYMBOL_MAP[symbol]


def from_bokeh_symbol(symbol: str, angle: float) -> Symbol:
    return _SYMBOL_MAP_INV[(symbol, angle)]


def to_bokeh_hatch(pattern: Hatch) -> str:
    return _HATCH_MAP[pattern]


def from_bokeh_hatch(pattern: str) -> Hatch:
    return _HATCH_MAP_INV[pattern]


# the second parameter is "angle" / pi * 2
_SYMBOL_MAP = {
    Symbol.CIRCLE: ("circle", 0),
    Symbol.SQUARE: ("square", 0),
    Symbol.TRIANGLE_UP: ("triangle", 0),
    Symbol.TRIANGLE_DOWN: ("inverted_triangle", 0),
    Symbol.TRIANGLE_LEFT: ("triangle", 1),
    Symbol.TRIANGLE_RIGHT: ("triangle", -1),
    Symbol.DIAMOND: ("diamond", 0),
    Symbol.CROSS: ("x", 0),
    Symbol.PLUS: ("cross", 0),
    Symbol.STAR: ("star", 0),
    Symbol.VBAR: ("dash", 1),
    Symbol.HBAR: ("dash", 0),
}

_SYMBOL_MAP_INV = {v: k for k, v in _SYMBOL_MAP.items()}

_HATCH_MAP = {
    Hatch.SOLID: "blank",
    Hatch.HORIZONTAL: "horizontal_line",
    Hatch.VERTICAL: "vertical_line",
    Hatch.CROSS: "cross",
    Hatch.DIAGONAL_BACK: "right_diagonal_line",
    Hatch.DIAGONAL_FORWARD: "left_diagonal_line",
    Hatch.DIAGONAL_CROSS: "diagonal_cross",
    Hatch.DOTS: "dot",
}
_HATCH_MAP_INV = {v: k for k, v in _HATCH_MAP.items()}
_HATCH_MAP_INV[""] = Hatch.SOLID


def to_html(canvas) -> str:
    from bokeh.plotting import save

    with (
        tempfile.TemporaryDirectory() as dirname,
        warnings.catch_warnings(),
    ):
        warnings.filterwarnings("ignore", category=UserWarning)
        filename = Path(dirname) / "whitecanvas.html"
        save(canvas._grid_plot, filename, title="whitecanvas")
        text = filename.read_text()
    return text
